"""Run Mutation Trials."""

from __future__ import annotations

import concurrent.futures
import logging
import shutil
import time
from typing import TYPE_CHECKING, Callable
from zipfile import ZipFile

from click import style

from . import PoodleTrialRunError
from .data_types import Mutant, MutantTrial, MutantTrialResult, PoodleConfig, PoodleWork, TestingResults, TestingSummary
from .runners import command_line
from .util import dynamic_import, mutate_lines

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)

builtin_runners = {
    "command_line": command_line.runner,
}


def get_runner(config: PoodleConfig) -> Callable:
    """Retrieve runner callable given internal runner name or external runner python name."""
    logger.debug("Runner: %s", config.runner)

    if config.runner in builtin_runners:
        return builtin_runners[config.runner]

    return dynamic_import(config.runner)


def clean_run_each_source_folder(work: PoodleWork) -> dict[Path, MutantTrial]:
    """Run a trial on each source folder with no mutation."""
    return {folder: clean_run_trial(work, folder) for folder in work.config.source_folders}


def clean_run_trial(work: PoodleWork, folder: Path) -> MutantTrial:
    """Run a trial with no mutation."""
    start = time.time()
    work.echo(f"Testing clean run of folder '{folder}'...", nl=False)
    mutant_trial = run_mutant_trial(
        config=work.config,
        echo=work.echo,
        folder_zip=work.folder_zips[folder],
        mutant=Mutant(
            mutator_name="",
            source_folder=folder,
            source_file=None,
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            text="",
        ),
        run_id=work.next_num(),
        runner=work.runner,
        timeout=None,
    )
    if mutant_trial.result.found:  # not expected
        work.echo(style("FAILED", fg="red"))
        raise PoodleTrialRunError("Clean Run Failed", mutant_trial.result.reason_desc)

    work.echo("PASSED")
    logger.info("Elapsed Time %.2f s", time.time() - start)

    return mutant_trial


def run_mutant_trails(work: PoodleWork, mutants: list[Mutant], timeout: float) -> TestingResults:
    """Run the Mutant Trials and collect results.

    Report status as execution proceeds.
    """
    start = time.time()
    work.echo("Testing mutants")
    with concurrent.futures.ProcessPoolExecutor(max_workers=work.config.max_workers) as executor:
        try:
            futures = [
                executor.submit(
                    run_mutant_trial,
                    work.config,
                    work.echo,
                    work.folder_zips[mutant.source_folder],
                    mutant,
                    work.next_num(),
                    work.runner,
                    timeout,
                )
                for mutant in mutants
            ]

            summary = TestingSummary()
            summary.trials = len(mutants)
            for future in concurrent.futures.as_completed(futures):
                if future.cancelled():
                    work.echo("Canceled")
                else:
                    mutant_trial: MutantTrial = future.result()
                    summary += mutant_trial.result
                    work.echo(
                        f"COMPLETED {summary.tested:>4}/{summary.trials:<4}"
                        f"\tFOUND {summary.found:>4}"
                        f"\tNOT FOUND {summary.not_found:>4}"
                        f"\tTIMEOUT {summary.timeout:>4}"
                        f"\tERRORS {summary.errors:>4}",
                    )
        except KeyboardInterrupt:
            work.echo("Received Keyboard Interrupt.  Cancelling Remaining Trials.")
            executor.shutdown(wait=True, cancel_futures=True)
            raise

    logger.info("Elapsed Time %.2f s", time.time() - start)

    return TestingResults(
        mutant_trials=[future.result() for future in futures],
        summary=summary,
    )


def run_mutant_trial(  # noqa: PLR0913
    config: PoodleConfig,
    echo: Callable,
    folder_zip: Path,
    mutant: Mutant,
    run_id: str,
    runner: Callable,
    timeout: float | None,
) -> MutantTrial:
    """Run Trial for specified Mutant.

    Create a Run Folder.
    Unzip the zip file to the Run Folder.
    Apply Mutation.
    Call the Trial Runner.
    Delete the Run Folder.
    Return MutantTrial with result data.
    """
    start = time.time()
    logging.basicConfig(format=config.log_format, level=config.log_level)

    logger.debug(
        "SETUP: run_id=%s folder_zip=%s file=%s:%s text='%s'",
        run_id,
        folder_zip,
        mutant.source_file,
        mutant.lineno,
        mutant.text,
    )

    run_folder = config.work_folder / ("run-" + run_id)
    run_folder.mkdir()

    with ZipFile(folder_zip, "r") as zip_file:
        zip_file.extractall(run_folder)

    if mutant.source_file:
        target_file = run_folder / mutant.source_file
        file_lines = target_file.read_text("utf-8").splitlines(keepends=True)
        file_lines = mutate_lines(mutant, file_lines)
        target_file.write_text(data="".join(file_lines), encoding="utf-8")

    logger.debug("START: run_id=%s run_folder=%s", run_id, run_folder)

    result: MutantTrialResult = runner(
        config=config,
        echo=echo,
        run_folder=run_folder,
        mutant=mutant,
        timeout=timeout,
    )

    shutil.rmtree(run_folder)

    duration = time.time() - start
    logger.debug("END: run_id=%s - Elapsed Time %.2f s", run_id, duration)

    return MutantTrial(mutant=mutant, result=result, duration=duration)



"""Mutate Constant Values."""

from __future__ import annotations

import ast

from poodle.data_types import FileMutation, Mutator


class NumberMutator(ast.NodeVisitor, Mutator):
    """Mutate Numbers."""

    # Various Keywords:
    # complex   3j
    # int       3, 0o3 0x3, 0b11
    # float     3.0

    mutator_name = "Number"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_Constant(self, node: ast.Constant) -> None:
        """Increase and Decrease values."""
        if isinstance(node.value, bool):
            return
        if isinstance(node.value, int):
            self.mutants.append(self.create_file_mutation(node, str(node.value + 1)))
            self.mutants.append(self.create_file_mutation(node, str(node.value - 1)))
        if isinstance(node.value, complex):
            self.mutants.append(self.create_file_mutation(node, str(node.value + 1j)))
            self.mutants.append(self.create_file_mutation(node, str(node.value - 1j)))
        if isinstance(node.value, float):
            if node.value == 0:
                self.mutants.append(self.create_file_mutation(node, str(1.0)))
            else:
                self.mutants.append(self.create_file_mutation(node, str(node.value * 2)))
                self.mutants.append(self.create_file_mutation(node, str(node.value / 2)))


class StringMutator(ast.NodeVisitor, Mutator):
    """Mutate String."""

    mutator_name = "String"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.add_parent_attr(parsed_ast)
        self.visit(parsed_ast)
        return self.mutants

    def visit_Constant(self, node: ast.Constant) -> None:
        """Mutate String values."""
        if isinstance(node.parent, ast.Expr):  # type: ignore [attr-defined]
            return

        if isinstance(node.value, str):
            node.value = f"XX{node.value}XX"
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node)))


class KeywordMutator(ast.NodeVisitor, Mutator):
    """Mutate Keywords.

    continue, break, False, True, None
    """

    mutator_name = "Keyword"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.add_parent_attr(parsed_ast)
        self.visit(parsed_ast)
        return self.mutants

    def visit_Break(self, node: ast.Break) -> None:
        """Replace break with continue."""
        self.mutants.append(self.create_file_mutation(node, "continue"))

    def visit_Continue(self, node: ast.Continue) -> None:
        """Replace continue with break."""
        self.mutants.append(self.create_file_mutation(node, "break"))

    def visit_Constant(self, node: ast.Constant) -> None:
        """Replace True, False, and None."""
        if self.is_annotation(node):
            return

        if node.value is True:
            self.mutants.append(self.create_file_mutation(node, "False"))
        if node.value is False:
            self.mutants.append(self.create_file_mutation(node, "True"))
        if node.value is None:
            self.mutants.append(self.create_file_mutation(node, "' '"))



"""Data Types."""

from .data import (
    FileMutation,
    Mutant,
    MutantTrial,
    MutantTrialResult,
    PoodleConfig,
    PoodleSerialize,
    TestingResults,
    TestingSummary,
)
from .interfaces import Mutator
from .work import PoodleWork



import importlib
import logging
from io import BytesIO
from pathlib import Path
from unittest import mock

import pytest
from wcmatch import glob

from poodle import PoodleInputError, config


@pytest.fixture(autouse=True)
def _test_wrapper():
    importlib.reload(config)
    config.poodle_config = None
    yield
    importlib.reload(config)


@pytest.fixture()
def logging_mock():
    with mock.patch("poodle.config.logging") as logging_mock:
        yield logging_mock


@pytest.fixture()
def get_option_from_config():
    with mock.patch("poodle.config.get_option_from_config") as get_option_from_config:
        yield get_option_from_config


def test_defaults():
    importlib.reload(config)

    assert config.default_source_folders == [Path("src"), Path("lib")]

    assert config.default_log_format == "%(levelname)s [%(process)d] %(name)s.%(funcName)s:%(lineno)d - %(message)s"
    assert config.default_log_level == logging.WARN

    assert config.default_file_flags == glob.GLOBSTAR | glob.NODIR
    assert config.default_file_filters == ["test_*.py", "*_test.py", "poodle_config.py", "setup.py"]

    assert config.default_file_copy_flags == glob.GLOBSTAR | glob.NODIR
    assert config.default_file_copy_filters == ["__pycache__/**"]
    assert config.default_work_folder == Path(".poodle-temp")

    assert config.default_mutator_opts == {}

    assert config.default_min_timeout == 10
    assert config.default_timeout_multiplier == 10
    assert config.default_runner == "command_line"
    assert config.default_runner_opts == {}

    assert config.default_reporters == ["summary", "not_found"]
    assert config.default_reporter_opts == {}


class TestMaxWorkers:
    @mock.patch("poodle.config.os")
    def test_default_max_workers_affinity(self, os_mock):
        os_mock.sched_getaffinity.return_value = [1, 2, 3, 4, 5, 6]
        os_mock.cpu_count.return_value = 8
        assert config.default_max_workers() == 5
        os_mock.sched_getaffinity.assert_called_with(0)
        os_mock.cpu_count.assert_not_called()

    @mock.patch("poodle.config.os")
    def test_default_max_workers_cpu(self, os_mock):
        del os_mock.sched_getaffinity
        os_mock.cpu_count.return_value = 2
        assert config.default_max_workers() == 1
        os_mock.cpu_count.assert_called()

    @mock.patch("poodle.config.os")
    def test_default_max_workers_cpu_1(self, os_mock):
        del os_mock.sched_getaffinity
        os_mock.cpu_count.return_value = 1
        assert config.default_max_workers() == 1
        os_mock.cpu_count.assert_called()

    @mock.patch("poodle.config.os")
    def test_default_max_workers_neither(self, os_mock):
        del os_mock.sched_getaffinity
        os_mock.cpu_count.return_value = None
        assert config.default_max_workers() == 1
        os_mock.cpu_count.assert_called()


class TestBuildConfig:
    @mock.patch("poodle.config.os")
    @mock.patch("poodle.config.get_config_file_path")
    @mock.patch("poodle.config.get_config_file_data")
    @mock.patch("poodle.config.get_cmd_line_log_level")
    @mock.patch("poodle.config.get_source_folders")
    @mock.patch("poodle.config.get_str_from_config")
    @mock.patch("poodle.config.get_str_list_from_config")
    @mock.patch("poodle.config.get_path_from_config")
    @mock.patch("poodle.config.get_dict_from_config")
    @mock.patch("poodle.config.get_bool_from_config")
    @mock.patch("poodle.config.get_int_from_config")
    @mock.patch("poodle.config.get_any_from_config")
    @mock.patch("poodle.config.get_any_list_from_config")
    @mock.patch("poodle.config.get_cmd_line_echo_enabled")
    def test_build_config(
        self,
        get_cmd_line_echo_enabled,
        get_any_list_from_config,
        get_any_from_config,
        get_int_from_config,
        get_bool_from_config,
        get_dict_from_config,
        get_path_from_config,
        get_str_list_from_config,
        get_str_from_config,
        get_source_folders,
        get_cmd_line_log_level,
        get_config_file_data,
        get_config_file_path,
        mock_os,
        logging_mock,
    ):
        get_dict_from_config.side_effect = [
            {"mutator": "value"},
            {"runner": "value"},
            {"reporter": "value"},
        ]

        cmd_sources = (Path("src"),)
        cmd_config_file = Path("config.toml")

        config_file_path = get_config_file_path.return_value
        config_file_data = get_config_file_data.return_value

        mock_os.sched_getaffinity.return_value = [1, 2, 3, 4, 5, 6]

        assert config.build_config(
            cmd_sources,
            cmd_config_file,
            cmd_quiet=1,
            cmd_verbose=2,
            cmd_max_workers=3,
            cmd_excludes=("notcov.py",),
            cmd_only_files=("example.py",),
        ) == config.PoodleConfig(
            config_file=config_file_path,
            source_folders=get_source_folders.return_value,
            only_files=get_str_list_from_config.return_value,
            file_flags=get_int_from_config.return_value,
            file_filters=get_str_list_from_config.return_value.__iadd__.return_value,
            file_copy_flags=get_int_from_config.return_value,
            file_copy_filters=get_str_list_from_config.return_value,
            work_folder=get_path_from_config.return_value,
            max_workers=get_int_from_config.return_value,
            log_format=get_str_from_config.return_value,
            log_level=get_any_from_config.return_value,
            echo_enabled=get_bool_from_config.return_value,
            echo_no_color=get_bool_from_config.return_value,
            mutator_opts={"mutator": "value"},
            skip_mutators=get_str_list_from_config.return_value,
            add_mutators=get_any_list_from_config.return_value,
            min_timeout=get_int_from_config.return_value,
            timeout_multiplier=get_int_from_config.return_value,
            runner=get_str_from_config.return_value,
            runner_opts={"runner": "value"},
            reporters=get_str_list_from_config.return_value,
            reporter_opts={"reporter": "value"},
        )

        # return PoodleConfig
        # source_folders
        get_source_folders.assert_called_once_with(cmd_sources, config_file_data)

        # only_files
        get_str_list_from_config.assert_any_call(
            "only_files",
            config_file_data,
            default=[],
            command_line=("example.py",),
        )
        # file_flags
        get_int_from_config.assert_any_call("file_flags", config_file_data, default=config.default_file_flags)
        # file_filters
        get_str_list_from_config.assert_any_call("file_filters", config_file_data, default=config.default_file_filters)
        get_str_list_from_config.return_value.__iadd__.assert_any_call(("notcov.py",))

        # file_copy_flags
        get_int_from_config.assert_any_call("file_copy_flags", config_file_data, default=config.default_file_copy_flags)
        # file_copy_filters
        get_str_list_from_config.assert_any_call(
            "file_copy_filters",
            config_file_data,
            default=config.default_file_copy_filters,
        )
        # work_folder
        get_path_from_config.assert_any_call("work_folder", config_file_data, default=config.default_work_folder)

        # max_workers
        get_int_from_config.assert_any_call("max_workers", config_file_data, default=5, command_line=3)

        # log_format
        get_str_from_config.assert_any_call("log_format", config_file_data, default=config.default_log_format)
        # log_level
        get_any_from_config.assert_any_call(
            "log_level",
            config_file_data,
            default=config.default_log_level,
            command_line=get_cmd_line_log_level.return_value,
        )
        get_cmd_line_log_level.assert_called_with(1, 2)
        logging_mock.basicConfig.assert_called_once_with(
            format=get_str_from_config.return_value,
            level=get_any_from_config.return_value,
        )

        # echo_enabled
        get_bool_from_config.assert_any_call(
            "echo_enabled",
            config_file_data,
            default=True,
            command_line=get_cmd_line_echo_enabled.return_value,
        )
        get_cmd_line_echo_enabled.assert_called_once_with(1)

        # echo_no_color
        get_bool_from_config.assert_any_call("echo_no_color", config_file_data)

        # mutator_opts
        get_dict_from_config.assert_any_call("mutator_opts", config_file_data, default=config.default_mutator_opts)

        # skip_mutators
        get_str_list_from_config.assert_any_call("skip_mutators", config_file_data, default=[])

        # add_mutators
        get_any_list_from_config.assert_any_call("add_mutators", config_file_data)

        # min_timeout
        get_int_from_config.assert_any_call("min_timeout", config_file_data)

        # timeout_multiplier
        get_int_from_config.assert_any_call("timeout_multiplier", config_file_data)

        # runner
        get_str_from_config.assert_any_call("runner", config_file_data, default=config.default_runner)

        # runner_opts
        get_dict_from_config.assert_any_call("runner_opts", config_file_data, default=config.default_runner_opts)

        # reporters
        get_str_list_from_config.assert_any_call("reporters", config_file_data, default=config.default_reporters)

        # reporter_opts
        get_dict_from_config.assert_any_call("reporter_opts", config_file_data, default=config.default_reporter_opts)

    @mock.patch("poodle.config.get_config_file_data")
    def test_build_config_defaults(self, get_config_file_data):
        get_config_file_data.return_value = {}

        assert config.build_config(
            cmd_sources=(),
            cmd_config_file=None,
            cmd_quiet=0,
            cmd_verbose=0,
            cmd_max_workers=None,
            cmd_excludes=(),
            cmd_only_files=(),
        ) == config.PoodleConfig(
            config_file=Path("pyproject.toml"),
            source_folders=[Path("src")],
            only_files=[],
            file_flags=config.default_file_flags,
            file_filters=config.default_file_filters,
            file_copy_flags=config.default_file_copy_flags,
            file_copy_filters=config.default_file_copy_filters,
            work_folder=Path(".poodle-temp"),
            max_workers=config.default_max_workers(),
            log_format=config.default_log_format,
            log_level=logging.WARN,
            echo_enabled=True,
            echo_no_color=None,
            mutator_opts={},
            skip_mutators=[],
            add_mutators=[],
            min_timeout=10,
            timeout_multiplier=10,
            runner="command_line",
            runner_opts={},
            reporters=["summary", "not_found"],
            reporter_opts={},
        )


class TestGetCommandLineLoggingOptions:
    @pytest.mark.parametrize(
        ("cmd_quiet", "cmd_verbose", "expected"),
        [
            (0, 0, None),
            (1, 0, logging.WARN),
            (2, 0, logging.ERROR),
            (3, 0, logging.CRITICAL),
            (4, 0, logging.CRITICAL),
            (1, 1, logging.WARN),
            (0, 1, logging.INFO),
            (0, 2, logging.DEBUG),
            (0, 3, logging.NOTSET),
            (0, 4, logging.NOTSET),
        ],
    )
    def test_get_cmd_line_log_level(self, cmd_quiet, cmd_verbose, expected):
        assert config.get_cmd_line_log_level(cmd_quiet, cmd_verbose) == expected

    @pytest.mark.parametrize(
        ("cmd_quiet", "expected"),
        [
            (0, None),
            (1, False),
            (2, False),
        ],
    )
    def test_get_cmd_line_echo_enabled(self, cmd_quiet, expected):
        assert config.get_cmd_line_echo_enabled(cmd_quiet) is expected


class TestGetConfigFilePath:
    def setup_mock_path(self, mock_path, config_is_file, poodle_is_file, pyproject_is_file):
        mock_path.config_file_path.is_file.return_value = config_is_file
        mock_path.poodle_toml_path.is_file.return_value = poodle_is_file
        mock_path.pyproject_toml_path.is_file.return_value = pyproject_is_file

        def get_path(filename):
            if filename == "config.toml":
                return mock_path.config_file_path
            if filename == "poodle.toml":
                return mock_path.poodle_toml_path
            if filename == "pyproject.toml":
                return mock_path.pyproject_toml_path
            return None

        mock_path.side_effect = get_path

    @mock.patch("poodle.config.Path")
    def test_get_config_file_path_cmd_line(self, mock_path):
        self.setup_mock_path(mock_path, True, True, True)
        path_config = mock.MagicMock()
        path_config.is_file.return_value = True
        assert config.get_config_file_path(path_config) == path_config

    @mock.patch("poodle.config.Path")
    def test_get_config_file_path_cmd_line_not_found(self, mock_path):
        self.setup_mock_path(mock_path, False, True, True)
        path_config = mock.MagicMock()
        path_config.is_file.return_value = False
        path_config.__repr__ = lambda _: "config.toml"
        with pytest.raises(PoodleInputError, match="^Config file not found: 'config.toml'$"):
            config.get_config_file_path(path_config)

    @mock.patch("poodle.config.Path")
    @mock.patch("poodle.config.poodle_config")
    def test_get_config_file_path_poodle_config(self, poodle_config, mock_path):
        self.setup_mock_path(mock_path, True, False, False)
        poodle_config.config_file = "config.toml"
        assert config.get_config_file_path(None) == mock_path.config_file_path

    @mock.patch("poodle.config.Path")
    @mock.patch("poodle.config.poodle_config")
    def test_get_config_file_path_poodle_config_err(self, poodle_config, mock_path):
        self.setup_mock_path(mock_path, False, False, False)
        poodle_config.config_file = "config.toml"
        mock_path.config_file_path.__repr__ = lambda _: "config.toml"
        with pytest.raises(PoodleInputError, match="^config_file not found: 'config.toml'$"):
            config.get_config_file_path(None)

    @mock.patch("poodle.config.Path")
    @mock.patch("poodle.config.poodle_config")
    def test_get_config_file_path_poodle(self, poodle_config, mock_path):
        del poodle_config.config_file
        self.setup_mock_path(mock_path, False, True, True)
        assert config.get_config_file_path(None) == mock_path.poodle_toml_path

    @mock.patch("poodle.config.Path")
    @mock.patch("poodle.config.poodle_config")
    def test_get_config_file_path_pyproject(self, poodle_config, mock_path):
        del poodle_config.config_file
        self.setup_mock_path(mock_path, False, False, True)
        assert config.get_config_file_path(None) == mock_path.pyproject_toml_path

    @mock.patch("poodle.config.Path")
    @mock.patch("poodle.config.poodle_config")
    def test_get_config_file_path_none(self, poodle_config, mock_path):
        del poodle_config.config_file
        self.setup_mock_path(mock_path, False, False, False)
        assert config.get_config_file_path(None) is None


class TestGetConfigFileData:
    def test_get_config_file_data_no_file(self):
        assert config.get_config_file_data(None) == {}

    @mock.patch("poodle.config.get_config_file_data_toml")
    def test_get_config_file_data(self, get_config_file_data_toml):
        get_config_file_data_toml.return_value = {"test": "value"}
        assert config.get_config_file_data(Path("config.toml")) == {"test": "value"}
        get_config_file_data_toml.assert_called_with(Path("config.toml"))

    def test_get_config_file_data_invalid(self):
        with pytest.raises(PoodleInputError, match="^Config file type not supported: --config_file='config.txt'$"):
            config.get_config_file_data(Path("config.txt"))


config_toml_poodle = """
[poodle]
test = "value1"
"""

config_toml_tool_poodle = """
[tool.poodle]
test = "value2"
"""

config_toml_no_data = """
[project]
test = "value3"
"""


class TestGetConfigFileDataToml:
    def test_get_config_file_data_toml_poodle(self):
        file_path = mock.MagicMock()
        file_path.open.return_value = BytesIO(bytes(config_toml_poodle, encoding="utf-8"))
        assert config.get_config_file_data_toml(file_path) == {"test": "value1"}
        file_path.open.assert_called_with(mode="rb")

    def test_get_config_file_data_toml_tool_poodle(self):
        file_path = mock.MagicMock()
        file_path.open.return_value = BytesIO(bytes(config_toml_tool_poodle, encoding="utf-8"))
        assert config.get_config_file_data_toml(file_path) == {"test": "value2"}
        file_path.open.assert_called_with(mode="rb")

    def test_get_config_file_data_toml_no_data(self):
        file_path = mock.MagicMock()
        file_path.open.return_value = BytesIO(bytes(config_toml_no_data, encoding="utf-8"))
        assert config.get_config_file_data_toml(file_path) == {}
        file_path.open.assert_called_with(mode="rb")


class TestGetSourceFolders:
    @mock.patch("poodle.config.get_path_list_from_config")
    def test_get_source_folders(self, get_path_list_from_config):
        path_project = mock.MagicMock()
        path_project.is_dir.return_value = True

        get_path_list_from_config.return_value = [path_project]

        path_src = mock.MagicMock()
        path_src.is_dir.return_value = False
        path_lib = mock.MagicMock()
        path_lib.is_dir.return_value = True
        config.default_source_folders = [path_src, path_lib]

        assert config.get_source_folders((Path("src"),), {"source_folders": []}) == [path_project]

        get_path_list_from_config.assert_called_with(
            option_name="source_folders",
            config_data={"source_folders": []},
            command_line=(Path("src"),),
            default=[path_lib],
        )

        path_project.is_dir.assert_called()
        path_src.is_dir.assert_called()
        path_lib.is_dir.assert_called()

    @mock.patch("poodle.config.get_path_list_from_config")
    def test_get_source_folders_not_found(self, get_path_list_from_config):
        get_path_list_from_config.return_value = []
        with pytest.raises(PoodleInputError, match="^No source folder found to mutate.$"):
            config.get_source_folders((), {})

    @mock.patch("poodle.config.get_path_list_from_config")
    def test_get_source_folders_not_folder(self, get_path_list_from_config):
        path_project = mock.MagicMock()
        path_project.is_dir.return_value = False
        path_project.__repr__ = lambda _: "project"

        get_path_list_from_config.return_value = [path_project]

        with pytest.raises(PoodleInputError, match="Source 'project' must be a folder."):
            config.get_source_folders((), {})


class TestGetBoolFromConfig:
    @pytest.mark.parametrize(("expected"), [(True), (False)])
    def test_default(self, expected, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_bool_from_config(
                option_name="test_option",
                config_data={"test_option": not expected},
                command_line=not expected,
                default=expected,
            )
            == expected
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": not expected},
            command_line=not expected,
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_bool_from_config(
                option_name="test_option",
                config_data={"test_option": True},
            )
            is None
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": True},
            command_line=None,
        )

    @pytest.mark.parametrize(
        ("option", "default", "expected"),
        [
            (True, False, True),
            (False, True, False),
            ("true", False, True),
            ("False", True, False),
            (None, True, True),
            (None, False, False),
        ],
    )
    def test_values(self, option, default, expected, get_option_from_config):
        get_option_from_config.return_value = (option, None)

        assert (
            config.get_bool_from_config(
                option_name="test_option",
                config_data={},
                default=default,
            )
            == expected
        )


class TestGetPathFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert config.get_path_from_config(
            option_name="test_option",
            config_data={"test_option": Path("config_file_value")},
            command_line=Path("command_line_value"),
            default=Path("default_value"),
        ) == Path("default_value")

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": Path("config_file_value")},
            command_line=Path("command_line_value"),
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert config.get_path_from_config(
            option_name="test_option",
            config_data={"test_option": Path("config_file_value")},
            default=Path("default_value"),
        ) == Path("default_value")

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": Path("config_file_value")},
            command_line=None,
        )

    def test_string(self, get_option_from_config):
        get_option_from_config.return_value = ("return_value", "Source Name")

        assert config.get_path_from_config(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="command_line_value",
            default="default_value",
        ) == Path("return_value")

    def test_empty_string(self, get_option_from_config):
        get_option_from_config.return_value = ("", "Source Name")

        assert (
            config.get_path_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
                command_line="command_line_value",
                default=Path("default_value"),
            )
            == Path()
        )

    def test_not_path(self, get_option_from_config):
        get_option_from_config.return_value = (123, "Source Name")

        with pytest.raises(PoodleInputError, match=r"^test_option from Source Name must be a valid StrPath$"):
            config.get_path_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=Path("default_value"),
            )


class TestGetPathListFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={"test_option": [Path("config_file_value")]},
            command_line=[Path("command_line_value")],
            default=[Path("default_value")],
        ) == [Path("default_value")]

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": [Path("config_file_value")]},
            command_line=[Path("command_line_value")],
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert (
            config.get_path_list_from_config(
                option_name="test_option",
                config_data={"test_option": [Path("config_file_value")]},
            )
            == []
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": [Path("config_file_value")]},
            command_line=(),
        )

    def test_path(self, get_option_from_config):
        get_option_from_config.return_value = (Path("return_value"), "Source Name")
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=[Path("default_value")],
        ) == [Path("return_value")]

    def test_string(self, get_option_from_config):
        get_option_from_config.return_value = ("return_value", "Source Name")
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=[Path("default_value")],
        ) == [Path("return_value")]

    def test_string_empty(self, get_option_from_config):
        get_option_from_config.return_value = ("", "Source Name")
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=[Path("default_value")],
        ) == [Path()]

    def test_path_list(self, get_option_from_config):
        get_option_from_config.return_value = ((Path("return_value"), Path("other_value")), "Source Name")
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=[Path("default_value")],
        ) == [Path("return_value"), Path("other_value")]

    def test_string_list(self, get_option_from_config):
        get_option_from_config.return_value = (("return_value", "other_value"), "Source Name")
        assert config.get_path_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=[Path("default_value")],
        ) == [Path("return_value"), Path("other_value")]

    def test_empty_list(self, get_option_from_config):
        get_option_from_config.return_value = ([], "Source Name")
        assert (
            config.get_path_list_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=[Path("default_value")],
            )
            == []
        )

    def test_not_iterable(self, get_option_from_config):
        get_option_from_config.return_value = (123, "Source Name")
        with pytest.raises(
            PoodleInputError,
            match=r"^test_option from Source Name must be a valid Iterable\[StrPath\]$",
        ):
            config.get_path_list_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=[Path("default_value")],
            )

    def test_not_path(self, get_option_from_config):
        get_option_from_config.return_value = ([123], "Source Name")
        with pytest.raises(
            PoodleInputError,
            match=r"^test_option from Source Name must be a valid Iterable\[StrPath\]$",
        ):
            config.get_path_list_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=[Path("default_value")],
            )


class TestGetAnyFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_any_from_config(
                option_name="test_option",
                config_data={"test_option": 3},
                command_line="4",
                default=5.5,
            )
            == 5.5
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": 3},
            command_line="4",
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_any_from_config(
                option_name="test_option",
                config_data={"test_option": 3},
            )
            is None
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": 3},
            command_line=None,
        )

    def test_return_value(self, get_option_from_config):
        get_option_from_config.return_value = (False, "Source Name")

        assert (
            config.get_any_from_config(
                option_name="test_option",
                config_data={"test_option": "3"},
                command_line="4",
                default="5",
            )
            is False
        )


class TestGetAnyListFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert config.get_any_list_from_config(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=["command_line_value"],
            default=["default_value"],
        ) == ["default_value"]

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=["command_line_value"],
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert (
            config.get_any_list_from_config(
                option_name="test_option",
                config_data={"test_option": ["config_file_value"]},
            )
            == []
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=(),
        )

    def test_string(self, get_option_from_config):
        get_option_from_config.return_value = ("return_value", "Source Name")
        assert config.get_any_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["return_value"]

    def test_iterable(self, get_option_from_config):
        get_option_from_config.return_value = (iter(["return_value", "other_value"]), "Source Name")
        assert config.get_any_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["return_value", "other_value"]

    def test_tuple(self, get_option_from_config):
        get_option_from_config.return_value = (("return_value", "other_value"), "Source Name")
        assert config.get_any_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["return_value", "other_value"]

    def test_int(self, get_option_from_config):
        get_option_from_config.return_value = (3, "Source Name")
        assert config.get_any_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == [3]


class TestGetIntFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_int_from_config(
                option_name="test_option",
                config_data={"test_option": 3},
                command_line=4,
                default=5,
            )
            == 5
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": 3},
            command_line=4,
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_int_from_config(
                option_name="test_option",
                config_data={"test_option": 3},
            )
            is None
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": 3},
            command_line=None,
        )

    def test_str_to_int(self, get_option_from_config):
        get_option_from_config.return_value = ("5", "Source Name")

        assert (
            config.get_int_from_config(
                option_name="test_option",
                config_data={"test_option": "3"},
                command_line="4",
                default="5",
            )
            == 5
        )

    def test_convert_error(self, get_option_from_config):
        get_option_from_config.return_value = ("a", "Source Name")

        with pytest.raises(ValueError, match="^test_option from Source Name must be a valid int$"):
            config.get_int_from_config(
                option_name="test_option",
                config_data={"test_option": "3"},
                command_line="4",
                default="5",
            )


class TestGetStrFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_str_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
                command_line="command_line_value",
                default="default_value",
            )
            == "default_value"
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="command_line_value",
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)

        assert (
            config.get_str_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
            )
            == ""
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="",
        )

    def test_string(self, get_option_from_config):
        get_option_from_config.return_value = ("return_value", "Source Name")

        assert (
            config.get_str_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
                command_line="command_line_value",
                default="default_value",
            )
            == "return_value"
        )

    def test_empty_string(self, get_option_from_config):
        get_option_from_config.return_value = ("", "Source Name")

        assert (
            config.get_str_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
                command_line="command_line_value",
                default="default_value",
            )
            == ""
        )

    def test_convert(self, get_option_from_config):
        get_option_from_config.return_value = (Path("source.py"), "Source Name")

        assert (
            config.get_str_from_config(
                option_name="test_option",
                config_data={"test_option": "config_file_value"},
                command_line="command_line_value",
                default="default_value",
            )
            == "source.py"
        )


class TestGetStrListFromConfig:
    def test_default(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert config.get_str_list_from_config(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=["command_line_value"],
            default=["default_value"],
        ) == ["default_value"]

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=["command_line_value"],
        )

    def test_default_inputs(self, get_option_from_config):
        get_option_from_config.return_value = (None, None)
        assert (
            config.get_str_list_from_config(
                option_name="test_option",
                config_data={"test_option": ["config_file_value"]},
            )
            == []
        )

        get_option_from_config.assert_called_with(
            option_name="test_option",
            config_data={"test_option": ["config_file_value"]},
            command_line=(),
        )

    def test_string(self, get_option_from_config):
        get_option_from_config.return_value = ("return_value", "Source Name")
        assert config.get_str_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["return_value"]

    def test_string_empty(self, get_option_from_config):
        get_option_from_config.return_value = ("", "Source Name")
        assert config.get_str_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == [""]

    def test_string_list(self, get_option_from_config):
        get_option_from_config.return_value = (("return_value", "other_value"), "Source Name")
        assert config.get_str_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["return_value", "other_value"]

    def test_string_list_convert(self, get_option_from_config):
        get_option_from_config.return_value = ((Path("output.txt"),), "Source Name")
        assert config.get_str_list_from_config(
            option_name="test_option",
            config_data={},
            command_line=[],
            default=["default_value"],
        ) == ["output.txt"]

    def test_string_list_empty(self, get_option_from_config):
        get_option_from_config.return_value = ([], "Source Name")
        assert (
            config.get_str_list_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=["default_value"],
            )
            == []
        )

    def test_type_error(self, get_option_from_config):
        get_option_from_config.return_value = (123, "Source Name")
        with pytest.raises(PoodleInputError, match=r"^test_option from Source Name must be a valid Iterable\[str\]$"):
            config.get_str_list_from_config(
                option_name="test_option",
                config_data={},
                command_line=[],
                default=["default_value"],
            )


class TestGetOptionFromConfig:
    def test_get_command_line_value(self):
        assert config.get_option_from_config(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="command_line_value",
        ) == ("command_line_value", "Command Line")

    def test_get_poodle_config_value(self):
        config.poodle_config = mock.MagicMock(test_option="poodle_config_value")
        assert config.get_option_from_config(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="",
        ) == ("poodle_config_value", "poodle_config.py")

    def test_get_config_data_value(self):
        assert config.get_option_from_config(
            option_name="test_option",
            config_data={"test_option": "config_file_value"},
            command_line="",
        ) == ("config_file_value", "config file")

    def test_not_found(self):
        value, source = config.get_option_from_config(
            option_name="test_option",
            config_data={},
            command_line="",
        )
        assert value is None
        assert source is None

    def test_command_line_false(self):
        value, source = config.get_option_from_config(
            option_name="test_option",
            config_data={},
            command_line=False,
        )
        assert value is False
        assert source == "Command Line"


class TestGetDictFromConfig:
    def test_default_only(self):
        assert config.get_dict_from_config(option_name="mutator_opts", config_data={}) == {}

    def test_config_data(self):
        assert config.get_dict_from_config(
            option_name="mutator_opts",
            default={"bin_op_level": "std"},
            config_data={
                "mutator_opts": {
                    "bin_op_level": "min",
                    "config_file_option": "ABCD",
                },
                "runner_opts": {
                    "mutator_file_option": "QWERTY",
                },
            },
        ) == {
            "bin_op_level": "min",
            "config_file_option": "ABCD",
        }

    def test_config_data_invalid(self):
        with pytest.raises(PoodleInputError, match="^mutator_opts in config file must be a valid dict$"):
            config.get_dict_from_config(
                option_name="mutator_opts",
                default={"bin_op_level": "std"},
                config_data={"mutator_opts": "min"},
            )

    def test_poodle_config(self):
        config.poodle_config = mock.MagicMock(
            mutator_opts={
                "bin_op_level": "max",
                "poodle_config_option": "EFGH",
            },
        )
        assert config.get_dict_from_config(
            option_name="mutator_opts",
            default={"bin_op_level": "std"},
            config_data={
                "mutator_opts": {
                    "bin_op_level": "min",
                    "config_file_option": "ABCD",
                },
            },
            command_line={"cmd_option": "cmd_value"},
        ) == {
            "bin_op_level": "max",
            "config_file_option": "ABCD",
            "poodle_config_option": "EFGH",
            "cmd_option": "cmd_value",
        }

    def test_poodle_config_invalid(self):
        config.poodle_config = mock.MagicMock(mutator_opts="min")
        with pytest.raises(PoodleInputError, match="^poodle_config.mutator_opts must be a valid dict$"):
            config.get_dict_from_config(option_name="mutator_opts", default={"bin_op_level": "std"}, config_data={})



"""Containers for Work-in-progress data."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import IO, TYPE_CHECKING, Any, AnyStr, Callable

import click

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from .data import PoodleConfig
    from .interfaces import Mutator


class PoodleWork:
    """Work-in-progress data gathered into single structure for easy passing between functions."""

    def __init__(self, config: PoodleConfig) -> None:
        """Init from PoodleConfig."""
        self.config = config
        self.folder_zips: dict[Path, Path] = {}
        self.mutators: list[Mutator | Callable] = []
        self.runner: Callable = lambda *_, **__: None
        self.reporters: list[Callable] = []

        self._echo_wrapper = EchoWrapper(config.echo_enabled, config.echo_no_color)
        self.echo = self._echo_wrapper.echo

        def number_generator() -> Generator[int, Any, None]:
            i = 1
            while True:
                yield i
                i += 1

        self._num_gen = number_generator()

    def next_num(self) -> str:
        """Return the next value from Sequence as a string."""
        return str(next(self._num_gen))


@dataclass
class EchoWrapper:
    """Contains config options related to echo function."""

    echo_enabled: bool | None
    echo_no_color: bool | None

    def echo(  # noqa: PLR0913
        self,
        message: Any | None = None,  # noqa: ANN401
        file: IO[AnyStr] | None = None,
        nl: bool = True,
        err: bool = False,
        color: bool | None = None,
        **styles: dict,
    ) -> None:
        """Calls 'click.secho' with settings from config."""  # noqa: D401
        if self.echo_no_color:
            color = False
        if self.echo_enabled:
            click.secho(message, file, nl, err, color, **styles)



"""Poodle core run process."""

from __future__ import annotations

import logging
import shutil

import click

from . import PoodleInputError, PoodleTrialRunError
from .data_types import PoodleConfig, PoodleWork
from .mutate import create_mutants_for_all_mutators, initialize_mutators
from .report import generate_reporters
from .run import clean_run_each_source_folder, get_runner, run_mutant_trails
from .util import calc_timeout, create_temp_zips, create_unified_diff, pprint_str

logger = logging.getLogger(__name__)


def main_process(config: PoodleConfig) -> None:
    """Poodle core run process."""
    try:
        work = PoodleWork(config)  # sets logging defaults
        logger.info("\n%s", pprint_str(config))

        if config.work_folder.exists():
            logger.info("delete %s", config.work_folder)
            shutil.rmtree(config.work_folder)

        create_temp_zips(work)

        work.mutators = initialize_mutators(work)
        work.runner = get_runner(config)
        work.reporters = list(generate_reporters(config))

        mutants = create_mutants_for_all_mutators(work)
        work.echo(f"Identified {len(mutants)} mutants")

        clean_run_results = clean_run_each_source_folder(work)
        timeout = calc_timeout(config, clean_run_results)
        results = run_mutant_trails(work, mutants, timeout)

        for trial in results.mutant_trials:
            trial.mutant.unified_diff = create_unified_diff(trial.mutant)

        for reporter in work.reporters:
            reporter(config=config, echo=work.echo, testing_results=results)

        logger.info("delete %s", config.work_folder)
        shutil.rmtree(config.work_folder)
    except PoodleInputError as err:
        for arg in err.args:
            click.echo(arg)
    except PoodleTrialRunError as err:
        for arg in err.args:
            click.echo(arg)



"""Mutate Comparison Operators."""

from __future__ import annotations

import ast
import re
from copy import deepcopy
from typing import ClassVar

from poodle.data_types import FileMutation, Mutator


class ComparisonMutator(ast.NodeVisitor, Mutator):
    """Mutate Comparison Operators."""

    # https://docs.python.org/3/library/ast.html#ast.Compare
    # https://www.w3schools.com/python/python_operators.asp
    # ast.Eq      ==
    # ast.Lt      <
    # ast.LtE     <=
    # ast.Gt      >
    # ast.GtE     >=
    # ast.Is      is
    # ast.IsNot   is not
    # ast.In      in
    # ast.NotIn   not in
    # ast.Or      or
    # ast.And     and

    type_map: ClassVar[dict[type, list[type]]] = {
        ast.Eq: [ast.NotEq],
        ast.NotEq: [ast.Eq],
        ast.Lt: [ast.GtE, ast.LtE],
        ast.LtE: [ast.Gt, ast.Lt],
        ast.Gt: [ast.LtE, ast.GtE],
        ast.GtE: [ast.Lt, ast.Gt],
        ast.Is: [ast.IsNot],
        ast.IsNot: [ast.Is],
        ast.In: [ast.NotIn],
        ast.NotIn: [ast.In],
        ast.Or: [ast.And],
        ast.And: [ast.Or],
    }

    mutator_name = "Compare"

    def __init__(self, *args, **kwargs) -> None:
        """Initialize and read settings."""
        super().__init__(*args, **kwargs)
        self.mutants: list[FileMutation] = []

        self.filter_patterns: list[str] = [
            r"__name__ == '__main__'",
        ]

        self.filter_patterns += self.config.mutator_opts.get("compare_filters", [])

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all Comparisons and return created mutants."""
        self.mutants = []
        self.add_parent_attr(parsed_ast)
        self.visit(parsed_ast)
        return self.mutants

    def visit_Compare(self, node: ast.Compare) -> None:
        """Identify replacement Operations and create Mutants."""
        text = ast.unparse(node)
        for pattern in self.filter_patterns:
            if re.match(pattern, text):
                return

        for idx, op in enumerate(node.ops):
            for new_op in self.type_map[type(op)]:
                mut = deepcopy(node)
                mut.ops[idx] = new_op()
                self.mutants.append(self.create_file_mutation(node, ast.unparse(mut)))

    def visit_BoolOp(self, node: ast.BoolOp) -> None:
        """Identify replacement Operations and create Mutants."""
        if self.is_annotation(node) or type(node.op) not in self.type_map:
            return

        text = ast.unparse(node)
        for pattern in self.filter_patterns:
            if re.match(pattern, text):
                return

        for new_op in self.type_map[type(node.op)]:
            mut = deepcopy(node)
            mut.op = new_op()
            self.mutants.append(self.create_file_mutation(node, ast.unparse(mut)))



"""Mutate Functions and Calls."""

from __future__ import annotations

import ast
from copy import deepcopy

from poodle.data_types import FileMutation, Mutator


class FunctionCallMutator(ast.NodeVisitor, Mutator):
    """Mutate Function Calls."""

    mutator_name = "FuncCall"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_Call(self, node: ast.Call) -> None:
        """Replace Function calls with None."""
        self.mutants.append(self.create_file_mutation(node, "None"))


class DictArrayCallMutator(ast.NodeVisitor, Mutator):
    """Mutate Calls to Dict or Array."""

    mutator_name = "DictArray"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.add_parent_attr(parsed_ast)
        self.visit(parsed_ast)
        return self.mutants

    def visit_Subscript(self, node: ast.Subscript) -> None:
        """Replace Call to retrieve from Dict or Array with None."""
        if self.is_annotation(node):
            return

        self.mutants.append(self.create_file_mutation(node, "None"))


class LambdaReturnMutator(ast.NodeVisitor, Mutator):
    """Mutate Return from Lambdas."""

    mutator_name = "Lambda"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_Lambda(self, node: ast.Lambda) -> None:
        """Replace body of Lambda with None or empty string."""
        if isinstance(node.body, ast.Constant) and node.body.value is None:
            node.body = ast.Constant("")
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node)))
        else:
            node.body = ast.Constant(None)
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node)))


class ReturnMutator(ast.NodeVisitor, Mutator):
    """Mutate Return from Functions."""

    mutator_name = "Return"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_Return(self, node: ast.Return) -> None:
        """Replace return statements with return None or Return empty string."""
        if node.value is None:
            return
        if isinstance(node.value, ast.Constant) and node.value.value is None:
            node.value = ast.Constant("")
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node)))
        else:
            node.value = ast.Constant(None)
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node)))


class DecoratorMutator(ast.NodeVisitor, Mutator):
    """Mutate Decorators."""

    mutator_name = "Decorator"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all nodes and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        """Remove Decorators on Function Definitions."""
        if node.decorator_list:
            for idx in range(len(node.decorator_list)):
                new_node = deepcopy(node)
                new_node.decorator_list.pop(idx)
                self.mutants.append(self.create_file_mutation(node, ast.unparse(new_node)))



"""Various utility functions."""

from __future__ import annotations

import difflib
import json
import logging
from copy import deepcopy
from io import StringIO
from pprint import pprint
from typing import TYPE_CHECKING, Any
from zipfile import ZipFile

from wcmatch.pathlib import Path

if TYPE_CHECKING:
    import pathlib

    from .data_types import Mutant, MutantTrial, PoodleConfig, PoodleSerialize, PoodleWork

logger = logging.getLogger(__name__)


def files_list_for_folder(
    folder: pathlib.Path,
    match_glob: str,
    flags: int | None,
    filter_globs: list[str],
) -> list[pathlib.Path]:
    """Retrieve list of files in specified folder.

    Search recursively for files matching match_glob.
    Remove files matching any of the filter_globs values.
    """
    logger.debug("files_list_for_folder folder=%s, match_glob=%s filter_globs=%s", folder, match_glob, filter_globs)

    files: list[pathlib.Path] = list(Path(folder).rglob(match_glob, flags=flags, exclude=filter_globs))  # type: ignore [arg-type]

    logger.debug("files_list_for_folder results: folder=%s files=%s", folder, files)
    return files


def files_list_for_source_folders(work: PoodleWork) -> dict[pathlib.Path, list[pathlib.Path]]:
    """Build map of Folder to all files in folder to include in zips."""
    return {
        folder: files_list_for_folder(
            folder=folder,
            match_glob="*",
            flags=work.config.file_copy_flags,
            filter_globs=work.config.file_copy_filters,
        )
        for folder in work.config.source_folders
    }


def create_temp_zips(work: PoodleWork) -> None:
    """Create a temporary zip file for each folder in source_folders."""
    work.config.work_folder.mkdir(parents=True, exist_ok=True)
    for folder, files in files_list_for_source_folders(work).items():
        zip_file = work.config.work_folder / ("src-" + work.next_num() + ".zip")
        logger.info("Creating zip file: %s", zip_file)
        work.folder_zips[folder] = zip_file
        with ZipFile(zip_file, "w") as target_zip:
            for file in files:
                logger.info("Adding file: %s", file)
                target_zip.write(file)


def dynamic_import(object_to_import: str) -> Any:  # noqa: ANN401
    """Import an Object represented by provided python name."""
    logger.debug("Import object: %s", object_to_import)
    parts = object_to_import.split(".")
    module_str = ".".join(parts[:-1])
    obj_def = parts[-1]
    module = __import__(module_str, fromlist=[obj_def])
    return getattr(module, obj_def)


def pprint_str(obj: Any) -> str:  # noqa: ANN401
    """Pretty Print an object to a string."""
    out = StringIO()
    pprint(obj, stream=out, width=150)  # noqa: T203
    return out.getvalue()


def calc_timeout(config: PoodleConfig, clean_run_results: dict[pathlib.Path, MutantTrial]) -> float:
    """Determine timeout value to use in runner."""
    max_clean_run = max([trial.duration for trial in clean_run_results.values()])
    return max(float(max_clean_run) * config.timeout_multiplier, config.min_timeout)


def mutate_lines(mutant: Mutant, file_lines: list[str]) -> list[str]:
    """Apply mutation to list of lines from file."""
    mut_lines = deepcopy(file_lines)
    prefix = mut_lines[mutant.lineno - 1][: mutant.col_offset]
    suffix = mut_lines[mutant.end_lineno - 1][mutant.end_col_offset :]

    mut_lines[mutant.lineno - 1] = prefix + mutant.text + suffix
    for _ in range(mutant.lineno, mutant.end_lineno):
        mut_lines.pop(mutant.lineno)

    return mut_lines


def create_unified_diff(mutant: Mutant) -> str | None:
    """Add unified diff to mutant."""
    if mutant.source_file:
        file_lines = mutant.source_file.read_text("utf-8").splitlines(keepends=True)
        file_name = str(mutant.source_file.resolve())
        return "".join(
            difflib.unified_diff(
                a=file_lines,
                b=mutate_lines(mutant, file_lines),
                fromfile=file_name,
                tofile=f"[Mutant] {file_name}:{mutant.lineno}",
            )
        )
    return None


def to_json(obj: PoodleSerialize, indent: int | str | None = None) -> str:
    """Convert dataclass to json string."""
    return json.dumps(obj, indent=indent, default=lambda x: x.to_dict())


def from_json(data: str, datatype: type[PoodleSerialize]) -> PoodleSerialize:
    """Convert json string to dataclass."""
    return datatype(**json.loads(data, object_hook=datatype.from_dict))



"""Mutate Unary Operations."""

from __future__ import annotations

import ast

from poodle.data_types import FileMutation, Mutator


class UnaryOperationMutator(ast.NodeVisitor, Mutator):
    """Mutate Unary Operations."""

    # https://docs.python.org/3/library/ast.html#ast.UnaryOp
    # https://www.w3schools.com/python/python_operators.asp
    # ast.UAdd      +
    # ast.USub      -
    # ast.Not       not
    # ast.Invert    ~

    mutator_name = "UnaryOp"
    mutants: list[FileMutation]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit all Unary Operations and return created mutants."""
        self.mutants = []
        self.visit(parsed_ast)
        return self.mutants

    def visit_UnaryOp(self, node: ast.UnaryOp) -> None:
        """Create mutations."""
        if isinstance(node.op, ast.UAdd):
            self.mutants.append(
                self.create_file_mutation(node, ast.unparse(ast.UnaryOp(op=ast.USub(), operand=node.operand))),
            )
        if isinstance(node.op, ast.USub):
            self.mutants.append(
                self.create_file_mutation(node, ast.unparse(ast.UnaryOp(op=ast.UAdd(), operand=node.operand))),
            )
        if isinstance(node.op, (ast.Not, ast.Invert)):
            self.mutants.append(self.create_file_mutation(node, ast.unparse(node.operand)))



"""Shared Data Classes used by Poodle."""

from __future__ import annotations

from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass
class PoodleSerialize:
    """Base Class for Data Classes that need to be serialized to JSON."""

    @staticmethod
    def from_dict(d: dict[str, Any]) -> dict[str, Any]:
        """Correct fields in Dictionary for JSON deserialization."""
        return d

    def to_dict(self) -> dict[str, Any]:
        """Convert to Dictionary for JSON serialization."""
        return asdict(self)


@dataclass
class PoodleConfig:
    """Configuration options resolved from command line and config files."""

    config_file: Path | None
    source_folders: list[Path]

    only_files: list[str]
    file_flags: int | None
    file_filters: list[str]

    file_copy_flags: int | None
    file_copy_filters: list[str]
    work_folder: Path

    max_workers: int | None

    log_format: str
    log_level: int | str
    echo_enabled: bool | None
    echo_no_color: bool | None

    mutator_opts: dict
    skip_mutators: list[str]
    add_mutators: list[Any]

    min_timeout: int
    timeout_multiplier: int
    runner: str
    runner_opts: dict

    reporters: list[str]
    reporter_opts: dict


@dataclass
class FileMutation:
    """Mutation instructions for the current file."""

    mutator_name: str
    lineno: int
    col_offset: int
    end_lineno: int
    end_col_offset: int
    text: str


@dataclass
class Mutant(FileMutation, PoodleSerialize):
    """Mutation instructions for a specific file and folder."""

    source_folder: Path
    source_file: Path | None
    unified_diff: str | None = None

    @staticmethod
    def from_dict(d: dict[str, Any]) -> dict[str, Any]:
        """Correct fields in Dictionary for JSON deserialization."""
        if "source_folder" in d:
            d["source_folder"] = Path(d["source_folder"])
        if "source_file" in d and d["source_folder"] is not None:
            d["source_file"] = Path(d["source_file"])
        return d

    def to_dict(self) -> dict[str, Any]:
        """Convert to Dictionary for JSON serialization."""
        d = asdict(self)
        d["source_folder"] = str(self.source_folder)
        if isinstance(self.source_file, Path):
            d["source_file"] = str(self.source_file)
        return d


@dataclass
class MutantTrialResult(PoodleSerialize):
    """Mutation Trial Result for a Mutant."""

    found: bool
    reason_code: str
    reason_desc: str | None = None

    RC_FOUND = "mutant_found"
    RC_NOT_FOUND = "mutant_not_found"
    RC_TIMEOUT = "timeout"
    RC_INCOMPLETE = "incomplete"
    RC_OTHER = "other"


@dataclass
class MutantTrial(PoodleSerialize):
    """Trial Result for a Mutant."""

    mutant: Mutant
    result: MutantTrialResult
    duration: float

    @staticmethod
    def from_dict(d: dict[str, Any]) -> dict[str, Any]:
        """Correct fields in Dictionary for JSON deserialization."""
        if "mutant" in d:
            d["mutant"] = Mutant(**Mutant.from_dict(d["mutant"]))
        if "result" in d:
            d["result"] = MutantTrialResult(**d["result"])
        return d

    def to_dict(self) -> dict[str, Any]:
        """Convert to Dictionary for JSON serialization."""
        d = asdict(self)
        d["mutant"] = self.mutant.to_dict()
        d["result"] = self.result.to_dict()
        return d


@dataclass
class TestingSummary(PoodleSerialize):
    """Summary Statistics for a Test Run."""

    trials: int = 0
    tested: int = 0
    found: int = 0
    not_found: int = 0
    timeout: int = 0
    errors: int = 0
    success_rate: float = 0.0

    @property
    def coverage_display(self) -> str:
        """Return a formatted string for the coverage percentage."""
        return f"{self.success_rate * 100:.2f}%"

    def __iadd__(self, result: MutantTrialResult) -> Self:
        """Update Testing Summary with data from MutantTrialResult."""
        if isinstance(result, MutantTrialResult):
            self.tested += 1
            if result.found:
                self.found += 1
            elif result.reason_code == MutantTrialResult.RC_NOT_FOUND:
                self.not_found += 1
            elif result.reason_code == MutantTrialResult.RC_TIMEOUT:
                self.timeout += 1
            else:
                self.errors += 1

        if self.trials > 0:
            self.success_rate = self.found / self.trials

        return self

    @staticmethod
    def from_dict(d: dict[str, Any]) -> dict[str, Any]:
        """Correct fields in Dictionary for JSON deserialization."""
        d.pop("coverage_display", None)
        return d

    def to_dict(self) -> dict[str, Any]:
        """Convert to Dictionary for JSON serialization."""
        d = asdict(self)
        d["coverage_display"] = self.coverage_display
        return d


@dataclass
class TestingResults(PoodleSerialize):
    """Collection of all trials and summary statistics."""

    mutant_trials: list[MutantTrial]
    summary: TestingSummary

    @staticmethod
    def from_dict(d: dict[str, Any]) -> dict[str, Any]:
        """Correct fields in Dictionary for JSON deserialization."""
        if "mutant_trials" in d:
            d["mutant_trials"] = [MutantTrial(**MutantTrial.from_dict(trial)) for trial in d["mutant_trials"]]
        if "summary" in d:
            d["summary"] = TestingSummary(**TestingSummary.from_dict(d["summary"]))
        return d

    def to_dict(self) -> dict[str, Any]:
        """Convert to Dictionary for JSON serialization."""
        d = asdict(self)
        d["mutant_trials"] = [trial.to_dict() for trial in self.mutant_trials]
        d["summary"] = self.summary.to_dict()
        return d



"""Run mutation tests."""

from __future__ import annotations

import logging
import os
import shlex
import subprocess
from pathlib import Path
from subprocess import TimeoutExpired

from poodle.data_types import Mutant, MutantTrialResult, PoodleConfig
from poodle.util import pprint_str

logger = logging.getLogger(__name__)


def runner(
    config: PoodleConfig, run_folder: Path, mutant: Mutant, timeout: float | None, *_, **__
) -> MutantTrialResult:
    """Run test of mutant with command line command in subprocess."""
    logger.info("Running: run_folder=%s timeout=%s", run_folder, timeout)

    cwd = Path.cwd().resolve()
    run_cwd = run_folder.resolve() if mutant.source_folder.resolve() == cwd else cwd
    run_source_folder = run_folder.resolve() / mutant.source_folder

    run_env = os.environ.copy()
    python_path = os.pathsep.join(
        [
            str(run_source_folder),
            str(Path.cwd().resolve()),
            run_env.get("PYTHONPATH", ""),
        ],
    )
    update_env = {
        "PYTHONDONTWRITEBYTECODE": "1",
        "PYTHONPATH": python_path,
        "MUT_SOURCE_FILE": str(mutant.source_file),
        "MUT_LINENO": str(mutant.lineno),
        "MUT_END_LINENO": str(mutant.end_lineno),
        "MUT_COL_OFFSET": str(mutant.col_offset),
        "MUT_END_COL_OFFSET": str(mutant.end_col_offset),
        "MUT_TEXT": str(mutant.text),
    }
    if "command_line_env" in config.runner_opts:
        update_env.update(config.runner_opts["command_line_env"])
    run_env.update(update_env)

    logger.debug("update_env=%s", pprint_str(update_env))

    cmd: str = config.runner_opts.get("command_line", "pytest -x --assert=plain -o pythonpath='{PYTHONPATH}'")
    cmd = cmd.format(PYTHONPATH=python_path)
    logger.debug("command: %s", cmd)

    try:
        result = subprocess.run(
            shlex.split(cmd),  # noqa: S603
            cwd=run_cwd,
            env=run_env,
            capture_output=True,
            check=False,
            timeout=timeout,
        )
    except TimeoutExpired as te:
        return MutantTrialResult(
            found=False,
            reason_code=MutantTrialResult.RC_TIMEOUT,
            reason_desc=f"TimeoutExpired {te}",
        )

    if result.returncode == 1:
        return MutantTrialResult(
            found=True,
            reason_code=MutantTrialResult.RC_FOUND,
            reason_desc=result.stdout.decode("utf-8", errors="replace")  # nomut: String
            + "\n"
            + result.stderr.decode("utf-8", errors="replace"),  # nomut: String
        )
    if result.returncode == 0:
        return MutantTrialResult(
            found=False,
            reason_code=MutantTrialResult.RC_NOT_FOUND,
        )
    return MutantTrialResult(
        found=True,
        reason_code=MutantTrialResult.RC_OTHER,
        reason_desc=result.stdout.decode("utf-8", errors="replace")  # nomut: String
        + "\n"
        + result.stderr.decode("utf-8", errors="replace"),  # nomut: String
    )



"""Collection of Mutators."""

from .calls import DecoratorMutator, DictArrayCallMutator, FunctionCallMutator, LambdaReturnMutator, ReturnMutator
from .compare import ComparisonMutator
from .constant import KeywordMutator, NumberMutator, StringMutator
from .operators import AugAssignMutator, BinaryOperationMutator
from .unary_op import UnaryOperationMutator



"""Command Line Interface."""

from __future__ import annotations

import sys
from pathlib import Path

import click

from . import PoodleInputError, core
from .config import build_config

CONTEXT_SETTINGS = {
    "max_content_width": 120,
}


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("sources", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.option("-c", "config_file", help="Configuration File.", type=click.Path(exists=True, path_type=Path))
@click.option("-q", "quiet", help="Quiet mode: q, qq, or qqq", count=True)
@click.option("-v", "verbose", help="Verbose mode: v, vv, or vvv", count=True)
@click.option("-w", "workers", help="Maximum number of parallel workers.", type=int)
@click.option("--exclude", help="Add a glob exclude file filter. Multiple allowed.", multiple=True)
@click.option("--only", help="Glob pattern for files to mutate. Multiple allowed.", multiple=True)
def main(
    sources: tuple[Path],
    config_file: Path | None,
    quiet: int,
    verbose: int,
    workers: int | None,
    exclude: tuple[str],
    only: tuple[str],
) -> None:
    """Poodle Mutation Test Tool."""
    try:
        config = build_config(sources, config_file, quiet, verbose, workers, exclude, only)
    except PoodleInputError as err:
        click.echo(err.args)
        sys.exit(4)

    try:
        core.main_process(config)
    except KeyboardInterrupt:
        click.echo("Aborted due to Keyboard Interrupt!")
        sys.exit(2)


# pytest return codes
# Exit code 0: All tests were collected and passed successfully
# Exit code 1: Tests were collected and run but some of the tests failed
# Exit code 2: Test execution was interrupted by the user
# Exit code 3: Internal error happened while executing tests
# Exit code 4: pytest command line usage error
# Exit code 5: No tests were collected

# nomut: start
if __name__ == "__main__":
    main()



from unittest import mock

import pytest

from poodle import PoodleInputError, PoodleTrialRunError, core
from poodle.data_types import MutantTrial
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def logger_mock():
    with mock.patch("poodle.core.logger") as logger_mock:
        yield logger_mock


def test_logger():
    assert core.logger.name == "poodle.core"


class TestMain:
    @mock.patch("poodle.core.PoodleWork")
    @mock.patch("poodle.core.pprint_str")
    @mock.patch("poodle.core.shutil")
    @mock.patch("poodle.core.create_temp_zips")
    @mock.patch("poodle.core.initialize_mutators")
    @mock.patch("poodle.core.get_runner")
    @mock.patch("poodle.core.generate_reporters")
    @mock.patch("poodle.core.create_mutants_for_all_mutators")
    @mock.patch("poodle.core.clean_run_each_source_folder")
    @mock.patch("poodle.core.run_mutant_trails")
    @mock.patch("poodle.core.click")
    def test_main(
        self,
        core_click: mock.MagicMock,
        run_mutant_trails: mock.MagicMock,
        clean_run_each_source_folder: mock.MagicMock,
        create_mutants_for_all_mutators: mock.MagicMock,
        generate_reporters: mock.MagicMock,
        get_runner: mock.MagicMock,
        initialize_mutators: mock.MagicMock,
        create_temp_zips: mock.MagicMock,
        shutil: mock.MagicMock,
        pprint_str: mock.MagicMock,
        poodle_work_class: mock.MagicMock,
        logger_mock: mock.MagicMock,
    ):
        work_folder = mock.MagicMock()
        work_folder.exists.return_value = True
        config = PoodleConfigStub(work_folder=work_folder, echo_enabled=True, min_timeout=10, timeout_multiplier=10)

        reporter1 = mock.MagicMock()
        reporter2 = mock.MagicMock()
        generate_reporters.return_value = iter([reporter1, reporter2])

        mutant1 = mock.MagicMock()
        mutant2 = mock.MagicMock()
        create_mutants_for_all_mutators.return_value = [mutant1, mutant2]

        clean_run_each_source_folder.return_value = {"folder": MutantTrial(mutant=None, result=None, duration=1.0)}  # type: ignore [arg-type]

        core.main_process(config)

        poodle_work_class.assert_called_with(config)
        work = poodle_work_class.return_value

        pprint_str.assert_called_with(config)
        logger_mock.info.assert_has_calls(
            [
                mock.call("\n%s", pprint_str.return_value),
                mock.call("delete %s", work_folder),
            ]
        )
        assert logger_mock.info.call_count == 3

        work_folder.exists.assert_called()
        logger_mock.info.assert_any_call("delete %s", work_folder)

        shutil.rmtree.assert_called_with(work_folder)
        assert shutil.rmtree.call_count == 2

        create_temp_zips.assert_called_with(work)

        initialize_mutators.assert_called_with(work)
        assert work.mutators == initialize_mutators.return_value
        get_runner.assert_called_with(config)
        assert work.runner == get_runner.return_value
        generate_reporters.assert_called_with(config)
        assert work.reporters == [reporter1, reporter2]

        create_mutants_for_all_mutators.assert_called_with(work)
        work.echo.assert_called_with("Identified 2 mutants")

        clean_run_each_source_folder.assert_called_with(work)
        run_mutant_trails.assert_called_with(work, [mutant1, mutant2], 10.0)
        results = run_mutant_trails.return_value

        reporter1.assert_called_with(config=config, echo=work.echo, testing_results=results)
        reporter2.assert_called_with(config=config, echo=work.echo, testing_results=results)

        logger_mock.info.assert_any_call("delete %s", work_folder)

    @mock.patch("poodle.core.PoodleWork")
    @mock.patch("poodle.core.pprint_str")
    @mock.patch("poodle.core.shutil")
    @mock.patch("poodle.core.create_temp_zips")
    @mock.patch("poodle.core.initialize_mutators")
    @mock.patch("poodle.core.get_runner")
    @mock.patch("poodle.core.generate_reporters")
    @mock.patch("poodle.core.create_mutants_for_all_mutators")
    @mock.patch("poodle.core.clean_run_each_source_folder")
    @mock.patch("poodle.core.run_mutant_trails")
    @mock.patch("poodle.core.click")
    def test_main_not_exists(
        self,
        core_click: mock.MagicMock,
        run_mutant_trails: mock.MagicMock,
        clean_run_each_source_folder: mock.MagicMock,
        create_mutants_for_all_mutators: mock.MagicMock,
        generate_reporters: mock.MagicMock,
        get_runner: mock.MagicMock,
        initialize_mutators: mock.MagicMock,
        create_temp_zips: mock.MagicMock,
        shutil: mock.MagicMock,
        pprint_str: mock.MagicMock,
        poodle_work_class: mock.MagicMock,
        logger_mock: mock.MagicMock,
    ):
        work_folder = mock.MagicMock()
        work_folder.exists.return_value = False
        config = PoodleConfigStub(work_folder=work_folder, echo_enabled=True, min_timeout=10, timeout_multiplier=10)

        reporter1 = mock.MagicMock()
        reporter2 = mock.MagicMock()
        generate_reporters.return_value = iter([reporter1, reporter2])

        mutant1 = mock.MagicMock()
        mutant2 = mock.MagicMock()
        create_mutants_for_all_mutators.return_value = [mutant1, mutant2]

        clean_run_each_source_folder.return_value = {"folder": MutantTrial(mutant=None, result=None, duration=1.0)}  # type: ignore [arg-type]

        core.main_process(config)

        logger_mock.info.assert_has_calls(
            [
                mock.call("\n%s", pprint_str.return_value),
                mock.call("delete %s", work_folder),
            ]
        )
        assert logger_mock.info.call_count == 2

        shutil.rmtree.assert_called_with(work_folder)
        assert shutil.rmtree.call_count == 1

    @mock.patch("poodle.core.PoodleWork")
    @mock.patch("poodle.core.pprint_str")
    @mock.patch("poodle.core.shutil")
    @mock.patch("poodle.core.create_temp_zips")
    @mock.patch("poodle.core.initialize_mutators")
    @mock.patch("poodle.core.get_runner")
    @mock.patch("poodle.core.generate_reporters")
    @mock.patch("poodle.core.create_mutants_for_all_mutators")
    @mock.patch("poodle.core.clean_run_each_source_folder")
    @mock.patch("poodle.core.run_mutant_trails")
    @mock.patch("poodle.core.click")
    def test_input_error(
        self,
        core_click: mock.MagicMock,
        run_mutant_trails: mock.MagicMock,
        clean_run_each_source_folder: mock.MagicMock,
        create_mutants_for_all_mutators: mock.MagicMock,
        generate_reporters: mock.MagicMock,
        get_runner: mock.MagicMock,
        initialize_mutators: mock.MagicMock,
        create_temp_zips: mock.MagicMock,
        shutil: mock.MagicMock,
        pprint_str: mock.MagicMock,
        poodle_work_class: mock.MagicMock,
        logger_mock: mock.MagicMock,
    ):
        poodle_work_class.side_effect = [PoodleInputError("Input Error", "Bad Input")]

        core.main_process(PoodleConfigStub())

        core_click.echo.assert_any_call("Input Error")
        core_click.echo.assert_any_call("Bad Input")

    @mock.patch("poodle.core.PoodleWork")
    @mock.patch("poodle.core.pprint_str")
    @mock.patch("poodle.core.shutil")
    @mock.patch("poodle.core.create_temp_zips")
    @mock.patch("poodle.core.initialize_mutators")
    @mock.patch("poodle.core.get_runner")
    @mock.patch("poodle.core.generate_reporters")
    @mock.patch("poodle.core.create_mutants_for_all_mutators")
    @mock.patch("poodle.core.clean_run_each_source_folder")
    @mock.patch("poodle.core.run_mutant_trails")
    @mock.patch("poodle.core.click")
    def test_trial_error(
        self,
        core_click: mock.MagicMock,
        run_mutant_trails: mock.MagicMock,
        clean_run_each_source_folder: mock.MagicMock,
        create_mutants_for_all_mutators: mock.MagicMock,
        generate_reporters: mock.MagicMock,
        get_runner: mock.MagicMock,
        initialize_mutators: mock.MagicMock,
        create_temp_zips: mock.MagicMock,
        shutil: mock.MagicMock,
        pprint_str: mock.MagicMock,
        poodle_work_class: mock.MagicMock,
        logger_mock: mock.MagicMock,
    ):
        poodle_work_class.side_effect = [PoodleTrialRunError("Trial Error", "Execution Failed")]

        core.main_process(PoodleConfigStub())

        core_click.echo.assert_any_call("Trial Error")
        core_click.echo.assert_any_call("Execution Failed")



"""Abstract Classes and Model functions for Mutators, Runners, and Reporters."""

from __future__ import annotations

import ast
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable

from .data import FileMutation, Mutant, MutantTrialResult, PoodleConfig, TestingResults

if TYPE_CHECKING:
    from pathlib import Path


def create_mutations(  # type: ignore [empty-body]
    config: PoodleConfig,
    echo: Callable,
    parsed_ast: ast.Module,
    file_lines: list[str],
    *_,
    **__,
) -> list[FileMutation]:
    """Create a list of Mutants for the provided parsed Module.

    This will be called once with parsed ast for each Module.
    """


class Mutator(ABC):
    """Abstract class for Mutators."""

    def __init__(self, config: PoodleConfig, echo: Callable, *_, **__) -> None:
        """Initialize PoodleMutator."""
        self.config = config
        self.echo = echo

    mutator_name = ""

    @abstractmethod
    def create_mutations(self, parsed_ast: ast.Module, file_lines: list[str], *_, **__) -> list[FileMutation]:
        """Create a list of Mutants for the provided parsed Module.

        This will be called once with parsed ast for each Module.
        """

    @classmethod
    def create_file_mutation(cls, node: ast.AST, text: str) -> FileMutation:
        """Create a FileMutation copying location data from specified node."""
        lineno, col_offset, end_lineno, end_col_offset = cls.get_location(node)

        return FileMutation(
            mutator_name=cls.mutator_name,
            lineno=lineno,
            col_offset=col_offset,
            end_lineno=end_lineno,
            end_col_offset=end_col_offset,
            text=text,
        )

    @staticmethod
    def get_location(node: ast.AST) -> tuple[int, int, int, int]:
        """Get location lines and columns that encompasses node and all child nodes."""
        lineno = node.lineno
        col_offset = node.col_offset
        end_lineno = node.end_lineno or node.lineno
        end_col_offset = node.end_col_offset or node.col_offset

        for n in ast.walk(node):
            if not hasattr(n, "lineno"):
                continue

            if n.lineno < lineno:
                lineno = n.lineno
                col_offset = n.col_offset
            elif n.lineno == lineno and n.col_offset < col_offset:
                col_offset = n.col_offset

            if not hasattr(n, "end_lineno") or not n.end_lineno:
                continue

            if n.end_lineno > end_lineno:
                end_lineno = n.end_lineno
                if n.end_col_offset:
                    end_col_offset = n.end_col_offset
            elif n.end_lineno == end_lineno and n.end_col_offset and n.end_col_offset > end_col_offset:
                end_col_offset = n.end_col_offset

        return (lineno, col_offset, end_lineno, end_col_offset)

    @staticmethod
    def add_parent_attr(parsed_ast: ast.Module) -> None:
        """Update all child nodes in tree with parent field."""
        for node in ast.walk(parsed_ast):
            for child in ast.iter_child_nodes(node):
                child.parent = node  # type: ignore [attr-defined]

    @classmethod
    def is_annotation(cls, node: ast.AST, child_node: ast.AST | None = None) -> bool:  # nomut: Keyword
        """Recursively search parent nodes to see if the starting node is part of an annotation.

        Returns true if a parent node points to this node in an annotation or returns attribute.
        """
        if not hasattr(node, "parent") or not node.parent:
            return False

        if hasattr(node, "annotation") and node.annotation == child_node:
            return True

        if hasattr(node, "returns") and node.returns == child_node:
            return True

        return cls.is_annotation(node.parent, child_node=node)


# runner method signature:
def runner(  # type: ignore [empty-body]
    config: PoodleConfig,
    echo: Callable,
    run_folder: Path,
    mutant: Mutant,
    timeout: float | None,
    *_,
    **__,
) -> MutantTrialResult:
    """Run trial of mutant in specified folder.

    Files from the source folder have been copied to the run folder, and mutation has been applied.
    """


# reporter method signature:
def reporter(  # type: ignore [empty-body]
    config: PoodleConfig,
    echo: Callable,
    testing_results: TestingResults,
    *_,
    **__,
) -> None:
    """Report on Testing Results."""



"""Report Testing Results."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable

from .reporters import report_json, report_not_found, report_summary
from .util import dynamic_import

if TYPE_CHECKING:
    from collections.abc import Generator

    from .data_types import PoodleConfig

logger = logging.getLogger(__name__)

builtin_reporters: dict[str, Callable] = {
    "summary": report_summary,
    "not_found": report_not_found,
    "json": report_json,
}


def generate_reporters(config: PoodleConfig) -> Generator[Callable, Any, None]:
    """Build list of reporter functions."""
    logger.debug("Reporters: %s", config.reporters)

    if config.reporters:
        for reporter_name in config.reporters:
            if reporter_name in builtin_reporters:
                yield builtin_reporters[reporter_name]
            else:
                yield dynamic_import(reporter_name)



import importlib
import re
from pathlib import Path
from unittest import mock

import pytest
from click.testing import CliRunner

from poodle import PoodleInputError, cli


class TestCli:
    @pytest.fixture(autouse=True)
    def _setup(self):
        importlib.reload(cli)

    @pytest.fixture()
    def main_process(self):
        with mock.patch("poodle.cli.core.main_process") as main_process:
            yield main_process

    @pytest.fixture()
    def build_config(self):
        with mock.patch("poodle.cli.build_config") as build_config:
            yield build_config

    @pytest.fixture()
    def echo(self):
        with mock.patch("poodle.cli.click.echo") as echo:
            yield echo

    @pytest.fixture()
    def runner(self):
        return CliRunner()

    def test_cli_help(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert result.output.find("Usage: main [OPTIONS] [SOURCES]...") is not None
        assert result.output.find("Poodle Mutation Test Tool.") is not None
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_cli_help_max_content_width(self, runner: CliRunner):
        with mock.patch("poodle.cli.click.command") as command:
            importlib.reload(cli)
            runner.invoke(cli.main, ["--help"])
        command.assert_called_with(context_settings={"max_content_width": 120})

    def test_cli_help_config_file(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-c PATH\s+Configuration File\..*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_quiet(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-q\s+Quiet mode: q, qq, or qqq.*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_verbose(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-v\s+Verbose mode: v, vv, or vvv.*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_workers(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*-w INTEGER\s+Maximum number of parallel workers\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_exclude(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--exclude TEXT\s+Add a glob exclude file filter\. Multiple allowed\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_only(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--only TEXT\s+Glob pattern for files to mutate\. Multiple allowed\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 0, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_sources(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["src", "tests"])
        assert result.exit_code == 0
        build_config.assert_called_with((Path("src"), Path("tests")), None, 0, 0, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_sources_not_exist(
        self,
        main_process: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        result = runner.invoke(cli.main, ["not_exist_path"])
        assert result.exit_code > 0
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_main_config_file(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-c", "pyproject.toml"])
        assert result.exit_code == 0
        build_config.assert_called_with((), Path("pyproject.toml"), 0, 0, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_config_file_not_exists(
        self,
        main_process: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        result = runner.invoke(cli.main, ["-c", "not_exist_path"])
        assert result.exit_code > 0
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_main_quiet(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-q"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 1, 0, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_quiet2(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-qq"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 2, 0, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_verbose(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-v"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 1, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_verbose2(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-vv"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 2, None, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_workers(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-w", "4"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 0, 4, (), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_exclude(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--exclude", "test_.*"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 0, None, ("test_.*",), ())
        main_process.assert_called_with(build_config.return_value)

    def test_main_only(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--only", "test_.*"])
        assert result.exit_code == 0
        build_config.assert_called_with((), None, 0, 0, None, (), ("test_.*",))
        main_process.assert_called_with(build_config.return_value)

    def test_main_input_error(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        build_config.side_effect = PoodleInputError("bad input")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 4
        build_config.assert_called()
        echo.assert_called_with(("bad input",))
        main_process.assert_not_called()

    def test_main_keyboard_interrupt(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = KeyboardInterrupt()
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 2
        build_config.assert_called()
        echo.assert_called_with("Aborted due to Keyboard Interrupt!")
        main_process.assert_called_with(build_config.return_value)



from concurrent.futures import Future
from pathlib import Path
from unittest import mock

import click
import pytest

from poodle import run
from poodle.data_types import Mutant, MutantTrial, MutantTrialResult, PoodleWork, TestingResults, TestingSummary
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def mock_logger():
    with mock.patch("poodle.run.logger") as logger:
        yield logger


@pytest.fixture()
def mock_echo():
    return mock.MagicMock()


@pytest.fixture()
def mock_time():
    with mock.patch("poodle.run.time") as time:
        yield time


def test_logger():
    assert run.logger.name == "poodle.run"


def test_builtin_runners():
    assert run.builtin_runners == {
        "command_line": run.command_line.runner,
    }


class TestGetRunner:
    def test_builtin_runner(self, mock_logger):
        config = PoodleConfigStub(runner="command_line")
        assert run.get_runner(config) is run.command_line.runner
        mock_logger.debug.assert_called_with("Runner: %s", config.runner)

    def test_external_runner(self, mock_logger):
        config = PoodleConfigStub(runner="poodle.runners.command_line.runner")
        assert run.get_runner(config) is run.command_line.runner
        mock_logger.debug.assert_called_with("Runner: %s", config.runner)

    def test_external_runner_not_found(self, mock_logger):
        config = PoodleConfigStub(runner="poodle.runners.not_found.runner")
        with pytest.raises(ImportError):
            run.get_runner(config)
        mock_logger.debug.assert_called_with("Runner: %s", config.runner)


class TestCleanRunTrial:
    @mock.patch("poodle.run.clean_run_trial")
    def test_clean_run_each_source_folder(self, clean_run_trial):
        clean_run_trial.side_effect = [1, 2, 3]
        folder_1 = Path("folder_1")
        folder_2 = Path("folder_2")
        folder_3 = Path("folder_3")

        work = PoodleWork(config=PoodleConfigStub(source_folders=[folder_1, folder_2, folder_3]))
        assert run.clean_run_each_source_folder(work) == {
            folder_1: 1,
            folder_2: 2,
            folder_3: 3,
        }

    @mock.patch("poodle.run.run_mutant_trial")
    def test_clean_run_trial(self, run_mutant_trial, mock_logger, mock_echo, mock_time):
        mock_time.time.side_effect = [1, 3]

        folder = Path("source_folder")

        work = PoodleWork(config=PoodleConfigStub())
        work.echo = mock_echo
        work.folder_zips = {folder: Path("folder.zip")}

        mutant = Mutant(
            mutator_name="",
            source_folder=folder,
            source_file=None,
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            text="",
        )

        trial = MutantTrial(mutant, MutantTrialResult(False, MutantTrialResult.RC_NOT_FOUND), 1)

        run_mutant_trial.return_value = trial

        assert run.clean_run_trial(work, folder) == trial

        assert run_mutant_trial.call_args.kwargs["config"] == work.config
        assert run_mutant_trial.call_args.kwargs["echo"] == mock_echo
        assert run_mutant_trial.call_args.kwargs["folder_zip"] == Path("folder.zip")
        assert run_mutant_trial.call_args.kwargs["mutant"] == mutant
        assert run_mutant_trial.call_args.kwargs["run_id"] == "1"
        assert run_mutant_trial.call_args.kwargs["runner"] == work.runner
        assert run_mutant_trial.call_args.kwargs["timeout"] is None

        mock_echo.assert_any_call(f"Testing clean run of folder '{folder}'...", nl=False)
        mock_echo.assert_any_call("PASSED")

        mock_logger.info.assert_called_with("Elapsed Time %.2f s", 2)

    @mock.patch("poodle.run.run_mutant_trial")
    def test_clean_run_trial_failed(self, run_mutant_trial, mock_echo):
        folder = Path("source_folder")

        work = PoodleWork(config=PoodleConfigStub())
        work.echo = mock_echo
        work.folder_zips = {folder: Path("folder.zip")}

        mutant = Mutant(
            mutator_name="",
            source_folder=folder,
            source_file=None,
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            text="",
        )

        trial = MutantTrial(mutant, MutantTrialResult(True, MutantTrialResult.RC_FOUND), 1)

        run_mutant_trial.return_value = trial

        with pytest.raises(run.PoodleTrialRunError) as err:  # noqa: PT012
            run.clean_run_trial(work, folder)
            assert err.value.args[0] == "Clean Run Failed"
            assert err.value.args[1] == MutantTrialResult.RC_FOUND

        mock_echo.assert_any_call(f"Testing clean run of folder '{folder}'...", nl=False)
        mock_echo.assert_any_call(click.style("FAILED", fg="red"))


class TestMutantTrials:
    def create_mutant(self, folder, text):
        return Mutant(
            mutator_name="",
            source_folder=folder,
            source_file=None,
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            text=text,
        )

    @mock.patch("poodle.run.concurrent")
    def test_run_mutant_trails(self, concurrent, mock_logger, mock_echo, mock_time):
        mock_time.time.side_effect = [1, 3]

        folder = Path("source_folder")

        work = PoodleWork(config=PoodleConfigStub(max_workers=10))
        work.echo = mock_echo
        work.folder_zips = {folder: Path("folder.zip")}

        mutants = [
            self.create_mutant(folder, "mut1"),
            self.create_mutant(folder, "mut2"),
        ]

        trials = [
            MutantTrial(mutants[0], MutantTrialResult(False, MutantTrialResult.RC_NOT_FOUND), 1),
            MutantTrial(mutants[1], MutantTrialResult(True, MutantTrialResult.RC_FOUND), 1),
        ]

        futures = [mock.MagicMock(spec=Future) for _ in trials]
        for i, trial in enumerate(trials):
            futures[i].result.return_value = trial
            futures[i].cancelled.return_value = False
        concurrent.futures.as_completed.return_value = futures

        executor = concurrent.futures.ProcessPoolExecutor.return_value.__enter__.return_value
        executor.submit.side_effect = futures

        summary = TestingSummary(trials=2, tested=2, found=1, not_found=1, success_rate=0.5)
        testing_results = TestingResults(mutant_trials=trials, summary=summary)

        actual_results = run.run_mutant_trails(work, mutants, 10)

        mock_echo.assert_any_call("Testing mutants")

        concurrent.futures.ProcessPoolExecutor.assert_called_with(max_workers=work.config.max_workers)

        for i, mutant in enumerate(mutants):
            executor.submit.assert_any_call(
                run.run_mutant_trial,
                work.config,
                mock_echo,
                Path("folder.zip"),
                mutant,
                str(i + 1),
                work.runner,
                10,
            )

        mock_echo.assert_any_call("COMPLETED    1/2   \tFOUND    0\tNOT FOUND    1\tTIMEOUT    0\tERRORS    0")
        mock_echo.assert_any_call("COMPLETED    2/2   \tFOUND    1\tNOT FOUND    1\tTIMEOUT    0\tERRORS    0")

        mock_logger.info.assert_called_with("Elapsed Time %.2f s", 2)

        assert actual_results == testing_results

    @mock.patch("poodle.run.concurrent")
    def test_run_mutant_trails_cancelled(self, concurrent, mock_echo, mock_time):
        mock_time.time.side_effect = [1, 3]

        folder = Path("source_folder")

        work = PoodleWork(config=PoodleConfigStub())
        work.echo = mock_echo
        work.folder_zips = {folder: Path("folder.zip")}

        mutants = [
            self.create_mutant(folder, "mut1"),
            self.create_mutant(folder, "mut2"),
        ]

        trials = [
            MutantTrial(mutants[0], MutantTrialResult(False, MutantTrialResult.RC_NOT_FOUND), 1),
            MutantTrial(mutants[1], MutantTrialResult(True, MutantTrialResult.RC_FOUND), 1),
        ]

        futures = [mock.MagicMock(spec=Future) for _ in trials]
        for i, trial in enumerate(trials):
            futures[i].result.return_value = trial
            futures[i].cancelled.return_value = True
        concurrent.futures.as_completed.return_value = futures

        executor = concurrent.futures.ProcessPoolExecutor.return_value.__enter__.return_value
        executor.submit.side_effect = futures

        summary = TestingSummary(trials=2, tested=0, found=0, not_found=0, success_rate=0)
        testing_results = TestingResults(mutant_trials=trials, summary=summary)

        actual_results = run.run_mutant_trails(work, mutants, 10)

        mock_echo.assert_has_calls(
            [
                mock.call("Testing mutants"),
                mock.call("Canceled"),
                mock.call("Canceled"),
            ]
        )

        assert actual_results == testing_results

    @mock.patch("poodle.run.concurrent")
    def test_run_mutant_trails_interrupt(self, concurrent, mock_echo):
        folder = Path("source_folder")

        work = PoodleWork(config=PoodleConfigStub())
        work.echo = mock_echo
        work.folder_zips = {folder: Path("folder.zip")}

        mutants = [
            self.create_mutant(folder, "mut1"),
            self.create_mutant(folder, "mut2"),
        ]

        executor = concurrent.futures.ProcessPoolExecutor.return_value.__enter__.return_value
        executor.submit.side_effect = KeyboardInterrupt()

        with pytest.raises(KeyboardInterrupt):
            run.run_mutant_trails(work, mutants, 10)

        mock_echo.assert_has_calls(
            [
                mock.call("Testing mutants"),
                mock.call("Received Keyboard Interrupt.  Cancelling Remaining Trials."),
            ]
        )

        executor.shutdown.assert_called_with(wait=True, cancel_futures=True)


class TestRunMutantTrial:
    def create_mutant(self, folder, source_file):
        return Mutant(
            mutator_name="",
            source_folder=folder,
            source_file=source_file,
            lineno=0,
            col_offset=0,
            end_lineno=0,
            end_col_offset=0,
            text="",
        )

    @mock.patch("poodle.run.logging")
    @mock.patch("poodle.run.ZipFile")
    @mock.patch("poodle.run.mutate_lines")
    @mock.patch("poodle.run.shutil")
    def test_run_mutant_trial(
        self,
        mock_shutil,
        mutate_lines,
        zip_file_cls,
        mock_logging,
        mock_logger,
        mock_echo,
        mock_time,
    ):
        mock_time.time.side_effect = [1, 3]
        mock_work_folder = mock.MagicMock(spec=Path)
        config = PoodleConfigStub(work_folder=mock_work_folder, log_format="log_format", log_level="DEBUG")
        runner = mock.MagicMock()
        folder = Path("folder")
        mutant = self.create_mutant(folder, Path("main.py"))

        run_folder = config.work_folder.__truediv__.return_value
        target_file = run_folder.__truediv__.return_value
        file_lines_orig = target_file.read_text.return_value.splitlines.return_value
        mutate_lines.return_value = ["line1\n", "line2\n"]

        result = runner.return_value

        returned_trial = run.run_mutant_trial(config, mock_echo, Path("folder.zip"), mutant, "1", runner, 10)

        mock_logging.basicConfig.assert_called_with(format=config.log_format, level=config.log_level)

        mock_logger.debug.assert_any_call(
            "SETUP: run_id=%s folder_zip=%s file=%s:%s text='%s'",
            "1",
            Path("folder.zip"),
            Path("main.py"),
            0,
            "",
        )

        config.work_folder.__truediv__.assert_called_with("run-1")
        run_folder.mkdir.assert_called_with()

        zip_file_cls.assert_called_with(Path("folder.zip"), "r")
        zip_file = zip_file_cls.return_value.__enter__.return_value
        zip_file.extractall.assert_called_with(run_folder)

        target_file.read_text.assert_called_with("utf-8")
        target_file.read_text.return_value.splitlines.assert_called_with(keepends=True)
        mutate_lines.assert_called_with(mutant, file_lines_orig)
        target_file.write_text.assert_called_with(data="line1\nline2\n", encoding="utf-8")

        mock_logger.debug.assert_any_call("START: run_id=%s run_folder=%s", "1", run_folder)

        runner.assert_called_with(
            config=config,
            echo=mock_echo,
            run_folder=run_folder,
            mutant=mutant,
            timeout=10,
        )

        mock_shutil.rmtree.assert_called_with(run_folder)

        mock_logger.debug.assert_any_call("END: run_id=%s - Elapsed Time %.2f s", "1", 2)

        assert returned_trial == MutantTrial(mutant, result, 2)

    @mock.patch("poodle.run.logging")
    @mock.patch("poodle.run.ZipFile")
    @mock.patch("poodle.run.mutate_lines")
    @mock.patch("poodle.run.shutil")
    def test_run_mutant_trial_no_source(
        self,
        mock_shutil,
        mutate_lines,
        zip_file_cls,
        mock_logging,
        mock_logger,
        mock_echo,
        mock_time,
    ):
        mock_time.time.side_effect = [1, 3]
        mock_work_folder = mock.MagicMock(spec=Path)
        config = PoodleConfigStub(work_folder=mock_work_folder, log_format="log_format", log_level="DEBUG")
        runner = mock.MagicMock()
        folder = Path("folder")
        mutant = self.create_mutant(folder, None)

        run_folder = config.work_folder.__truediv__.return_value
        target_file = run_folder.__truediv__.return_value

        result = runner.return_value

        returned_trial = run.run_mutant_trial(config, mock_echo, Path("folder.zip"), mutant, "1", runner, 10)

        mock_logging.basicConfig.assert_called_with(format=config.log_format, level=config.log_level)

        mock_logger.debug.assert_any_call(
            "SETUP: run_id=%s folder_zip=%s file=%s:%s text='%s'",
            "1",
            Path("folder.zip"),
            None,
            0,
            "",
        )

        config.work_folder.__truediv__.assert_called_with("run-1")
        run_folder.mkdir.assert_called_with()

        zip_file_cls.assert_called_with(Path("folder.zip"), "r")
        zip_file = zip_file_cls.return_value.__enter__.return_value
        zip_file.extractall.assert_called_with(run_folder)

        run_folder.__truediv__.assert_not_called()
        target_file.read_text.assert_not_called()
        mutate_lines.assert_not_called()
        target_file.write_text.assert_not_called()

        mock_logger.debug.assert_any_call("START: run_id=%s run_folder=%s", "1", run_folder)

        runner.assert_called_with(
            config=config,
            echo=mock_echo,
            run_folder=run_folder,
            mutant=mutant,
            timeout=10,
        )

        mock_shutil.rmtree.assert_called_with(run_folder)

        mock_logger.debug.assert_any_call("END: run_id=%s - Elapsed Time %.2f s", "1", 2)

        assert returned_trial == MutantTrial(mutant, result, 2)



from unittest import mock

import pytest

from poodle import PoodleInputError, mutate
from poodle.data_types import FileMutation, Mutant, Mutator, PoodleWork
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def logger_mock():
    with mock.patch("poodle.mutate.logger") as logger_mock:
        yield logger_mock


def file_mutation(mutator_name, lineno, end_lineno):
    return FileMutation(
        mutator_name=mutator_name,
        lineno=lineno,
        col_offset=0,
        end_lineno=end_lineno,
        end_col_offset=0,
        text="",
    )


class FakeMutator(Mutator):
    mutator_name = "FakeMutator"

    def create_mutations(self, *_, **__) -> list[FileMutation]:
        return []


def fake_mutator(*_, **__) -> list[FileMutation]:
    return []


class TestInit:
    @mock.patch("poodle.mutate.builtin_mutators")
    def test_initialize_mutators(self, builtin_mutators):
        config = PoodleConfigStub(skip_mutators=["m2"], add_mutators=["tests.test_mutate.fake_mutator"])
        work = PoodleWork(config)
        m1 = FakeMutator(config, echo=mock.MagicMock())
        m2 = FakeMutator(config, echo=mock.MagicMock())
        m3 = FakeMutator(config, echo=mock.MagicMock())
        builtin_mutators.items.return_value = [("m1", m1), ("m2", m2), ("m3", m3)]
        assert mutate.initialize_mutators(work) == [m1, m3, fake_mutator]

    def test_initialize_mutator_str_class(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        mutator = mutate.initialize_mutator(work, "tests.test_mutate.FakeMutator")
        logger_mock.debug.assert_called_with("tests.test_mutate.FakeMutator")
        assert isinstance(mutator, FakeMutator)
        assert not isinstance(mutator, type)
        assert mutator.config == config
        assert mutator.echo == work.echo

    def test_initialize_mutator_str_func(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        assert mutate.initialize_mutator(work, "tests.test_mutate.fake_mutator") == fake_mutator
        logger_mock.debug.assert_called_with("tests.test_mutate.fake_mutator")

    def test_initialize_mutator_class(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        mutator = mutate.initialize_mutator(work, FakeMutator)
        logger_mock.debug.assert_called_with(FakeMutator)
        assert isinstance(mutator, FakeMutator)
        assert not isinstance(mutator, type)
        assert mutator.config == config
        assert mutator.echo == work.echo

    def test_initialize_mutator_obj(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        mutator = FakeMutator(config, echo=mock.MagicMock())
        assert mutate.initialize_mutator(work, mutator) == mutator
        logger_mock.debug.assert_called_with(mutator)

    def test_initialize_mutator_func(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        assert mutate.initialize_mutator(work, fake_mutator) == fake_mutator
        logger_mock.debug.assert_called_with(fake_mutator)

    def test_initialize_mutator_invalid_error(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        with pytest.raises(
            PoodleInputError,
            match=(
                "^Unable to create mutator '3' of type=<class 'int'>. "
                "Expected String, Callable, Mutator subclass or Mutator subclass instance.$"
            ),
        ):
            mutate.initialize_mutator(work, 3)
        logger_mock.debug.assert_called_with(3)

    def test_initialize_mutator_import_error(self, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        with pytest.raises(
            PoodleInputError,
            match="^Import failed for mutator 'fake_mutator'$",
        ):
            mutate.initialize_mutator(work, "fake_mutator")
        logger_mock.debug.assert_called_with("fake_mutator")


class TestCreateMutants:
    @mock.patch("poodle.mutate.get_target_files")
    @mock.patch("poodle.mutate.create_mutants_for_file")
    def test_create_mutants_for_all_mutators(self, create_mutants_for_file, get_target_files):
        config = PoodleConfigStub()
        work = PoodleWork(config)

        get_target_files.return_value = {
            "example_1": ["file_1.py", "file_2.py"],
            "example_2": ["file_3.py", "file_4.py"],
        }

        def mock_create_mutants_for_file(work, folder, file):  # noqa: ARG001
            return [mock.MagicMock(folder=folder, file=file)]

        create_mutants_for_file.side_effect = mock_create_mutants_for_file

        out = mutate.create_mutants_for_all_mutators(work)

        get_target_files.assert_called_with(work)
        create_mutants_for_file.assert_has_calls(
            [
                mock.call(work, "example_1", "file_1.py"),
            ]
        )

        assert [(mock_out.folder, mock_out.file) for mock_out in out] == [
            ("example_1", "file_1.py"),
            ("example_1", "file_2.py"),
            ("example_2", "file_3.py"),
            ("example_2", "file_4.py"),
        ]

    @mock.patch("poodle.mutate.files_list_for_folder")
    def test_get_target_files_only(self, files_list_for_folder, logger_mock):
        files_list_for_folder.side_effect = [
            ["example1.py"],
            ["example2.py"],
            [],
            [],
        ]

        folder1 = mock.MagicMock()
        folder2 = mock.MagicMock()

        config = PoodleConfigStub(
            only_files=["example1.py", "example2.py"],
            file_flags=0,
            source_folders=[folder1, folder2],
        )
        work = PoodleWork(config)

        assert mutate.get_target_files(work) == {
            folder1: ["example1.py", "example2.py"],
            folder2: [],
        }

        logger_mock.debug.assert_called_with(
            "get_target_files source_folders=%s only_files=%s file_filters=%s",
            work.config.source_folders,
            work.config.only_files,
            work.config.source_folders,
        )

        files_list_for_folder.assert_has_calls(
            [
                mock.call(folder=folder1, match_glob="example1.py", flags=0, filter_globs=[]),
                mock.call(folder=folder1, match_glob="example2.py", flags=0, filter_globs=[]),
                mock.call(folder=folder2, match_glob="example1.py", flags=0, filter_globs=[]),
                mock.call(folder=folder2, match_glob="example2.py", flags=0, filter_globs=[]),
            ]
        )

    @mock.patch("poodle.mutate.files_list_for_folder")
    def test_get_target_files(self, files_list_for_folder):
        config = PoodleConfigStub(file_filters=["test.*.py"], file_flags=0, source_folders=["example_1", "example_2"])
        work = PoodleWork(config)

        files_list_for_folder.side_effect = [
            ["file_1.py", "file_2.py"],
            ["file_3.py", "file_4.py"],
        ]

        assert mutate.get_target_files(work) == {
            "example_1": ["file_1.py", "file_2.py"],
            "example_2": ["file_3.py", "file_4.py"],
        }

        files_list_for_folder.assert_has_calls(
            [
                mock.call(folder="example_1", match_glob="*.py", flags=0, filter_globs=["test.*.py"]),
                mock.call(folder="example_2", match_glob="*.py", flags=0, filter_globs=["test.*.py"]),
            ]
        )

    @mock.patch("poodle.mutate.ast")
    @mock.patch("poodle.mutate.deepcopy")
    def test_create_mutants_for_file(self, deepcopy, ast, logger_mock):
        config = PoodleConfigStub()
        work = PoodleWork(config)
        folder = mock.MagicMock()
        file = mock.MagicMock()

        parsed_ast = ast.parse.return_value
        file_lines = [
            "1 # nomut",
            "2",
        ]

        file.read_text.return_value.splitlines.return_value = file_lines

        def deepcopy_mock(obj):
            if obj == parsed_ast:
                return deepcopy.parsed_ast
            if obj == file_lines:
                return deepcopy.file_lines
            return deepcopy.other

        deepcopy.side_effect = deepcopy_mock

        mutator_1 = mock.MagicMock()
        file_mutants_1 = [file_mutation("Example", 1, 1), file_mutation("Example", 2, 2)]
        mutator_1.return_value = file_mutants_1

        mutator_2 = mock.MagicMock(spec=Mutator)
        file_mutants_2 = [file_mutation("Other", 1, 1), file_mutation("Other", 2, 2)]
        mutator_2.create_mutations.return_value = file_mutants_2

        work.mutators = [mutator_1, mutator_2]

        out_mutants = mutate.create_mutants_for_file(work, folder, file)

        logger_mock.debug.assert_called_with("Create Mutants for file %s", file)
        file.read_bytes.assert_called_once_with()
        file.read_text.assert_called_once_with("utf-8")
        ast.parse.assert_called_with(file.read_bytes.return_value, file)
        mutator_1.assert_called_with(config=config, parsed_ast=deepcopy.parsed_ast, file_lines=deepcopy.file_lines)
        mutator_2.create_mutations.assert_called_with(parsed_ast=deepcopy.parsed_ast, file_lines=deepcopy.file_lines)

        assert out_mutants == [
            Mutant(
                mutator_name="Example",
                lineno=2,
                col_offset=0,
                end_lineno=2,
                end_col_offset=0,
                text="",
                source_folder=folder,
                source_file=file,
            ),
            Mutant(
                mutator_name="Other",
                lineno=2,
                col_offset=0,
                end_lineno=2,
                end_col_offset=0,
                text="",
                source_folder=folder,
                source_file=file,
            ),
        ]


class TestFilter:
    def test_parse_filters(self):
        file_lines = [
            "1",
            "2 # pragma: no mutate",
            "3 # pragma: no mutate # type: ignore",
            "4 # nomut",
            "5 # nomut: Example",
            "6 # nomut: Example # type: ignore",
            "7 # nomut Example",
            "8 # nomut: Example1,Example2",
            "9 # nomut: Example1, Example2 # pragma: no mutate",
        ]
        assert mutate.parse_filters(file_lines) == {
            2: {"all"},
            3: {"all"},
            4: {"all"},
            5: {"example"},
            6: {"example"},
            7: {"example"},
            8: {"example1", "example2"},
            9: {"example1", "example2", "all"},
        }

    def test_parse_filters_start_end(self):
        file_lines = [
            "1",
            "2 # nomut: start",
            "3",
            "4 # nomut: end",
            "5",
            "6",
            "7",
            "8 # nomut: start",
            "9 # nomut: Example1, Example2",
        ]
        assert mutate.parse_filters(file_lines) == {
            2: {"all"},
            3: {"all"},
            4: {"all"},
            8: {"all"},
            9: {"all"},
        }

    def test_parse_filters_on_off(self):
        file_lines = [
            "1",
            "2 # nomut: on",
            "3",
            "4 # nomut: off",
            "5",
            "6",
            "7",
            "8 # nomut: on",
            "9 # nomut: Example1, Example2",
        ]
        assert mutate.parse_filters(file_lines) == {
            2: {"all"},
            3: {"all"},
            4: {"all"},
            8: {"all"},
            9: {"all"},
        }

    def test_add_filter(self):
        line_filters = {}
        mutate.add_line_filter(line_filters, 3, "all")
        assert line_filters == {3: {"all"}}
        mutate.add_line_filter(line_filters, 3, "Example")
        assert line_filters == {3: {"all", "example"}}
        mutate.add_line_filter(line_filters, 4, "Example")
        assert line_filters == {3: {"all", "example"}, 4: {"example"}}
        mutate.add_line_filter(line_filters, 5, "")
        assert line_filters == {3: {"all", "example"}, 4: {"example"}, 5: {"all"}}

    @pytest.mark.parametrize(
        ("mutation_args", "mutators", "expected"),
        [
            # all
            (("Example", 3, 3), {"all", "other"}, True),
            (("Example", 2, 3), {"all", "other"}, True),
            (("Example", 3, 4), {"all", "other"}, True),
            (("Example", 2, 4), {"all", "other"}, True),
            (("Example", 4, 5), {"all", "other"}, False),
            (("Example", 1, 2), {"all", "other"}, False),
            # match
            (("Example", 3, 3), {"example", "other"}, True),
            (("Example", 2, 3), {"example", "other"}, True),
            (("Example", 3, 4), {"example", "other"}, True),
            (("Example", 2, 4), {"example", "other"}, True),
            (("Example", 4, 5), {"example", "other"}, False),
            (("Example", 1, 2), {"example", "other"}, False),
            # no match
            (("Example", 3, 3), {"other", "other2"}, False),
            (("Example", 2, 3), {"other", "other2"}, False),
            (("Example", 3, 4), {"other", "other2"}, False),
            (("Example", 2, 4), {"other", "other2"}, False),
            (("Example", 4, 5), {"other", "other2"}, False),
            (("Example", 1, 2), {"other", "other2"}, False),
        ],
    )
    def test_is_filtered(self, mutation_args, mutators, expected):
        file_mutant = file_mutation(mutation_args[0], mutation_args[1], mutation_args[2])
        line_filters = {3: mutators}

        assert mutate.is_filtered(line_filters, file_mutant) is expected
